# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from oslo_config import cfg
from oslo_utils import importutils

from networking_fujitsu.ml2.fossw import fossw_vlandriver
from networking_fujitsu.ml2.fossw import mech_fossw
from networking_fujitsu.tests.unit.ml2.common import helper

from neutron.plugins.ml2.common import exceptions as ml2_exc
from neutron.plugins.ml2 import config as ml2_config
from neutron.tests.unit.plugins.ml2 import test_plugin as test_ml2_plugin

ADDRESS = '192.168.100.1'
ADDRESS2 = '192.168.100.2'
USERNAME = 'fossw_user'
PASSWORD = 'fossw_password'


class TestFujitsuMechDriverV2(test_ml2_plugin.Ml2PluginV2TestCase):
    """Test Fujitsu mechanism driver.

    """

    _mechanism_drivers = ['fujitsu_fossw']
    _skip = ["test_create_router_port_and_fail_create_postcommit"]

    def setUp(self):
        ml2_config.cfg.CONF.set_override(
            'tenant_network_types', ['vlan', 'vxlan'], 'ml2')

        if self._testMethodName in self._skip:
            self.skipTest("This test has already verified at neutron's test.")

        def mocked_initialize(self):
            self.ips = [ADDRESS, ADDRESS2]
            self.username = USERNAME
            self.password = PASSWORD
            self._vlan_driver = mock.Mock()
            self._vxlan_driver = mock.Mock()
            self.switches_mac_ip_pair = {
                "00:00:4c:ee:e5:39": ADDRESS,
                "00:00:4c:ee:e5:40": ADDRESS2,
            }
        with mock.patch.object(mech_fossw.FOSSWMechanismDriver,
                               'initialize', new=mocked_initialize):
            self.mech = mech_fossw.FOSSWMechanismDriver()
            super(TestFujitsuMechDriverV2, self).setUp()


class TestFujitsuMechDriverNetworksV2(test_ml2_plugin.TestMl2NetworksV2,
                                      TestFujitsuMechDriverV2):
    pass


class TestFujitsuMechDriverPortsV2(test_ml2_plugin.TestMl2PortsV2,
                                   TestFujitsuMechDriverV2):
    pass


class TestFOSSWInitialize(helper.FujitsuMechanismHelper):

    def setUp(self):
        super(TestFOSSWInitialize, self).setUp()
        ml2_config.cfg.CONF.set_override(
            'tenant_network_types', ['vlan', 'vxlan'], 'ml2')
        ml2_config.cfg.CONF.set_override(
            'username', 'admin', 'fujitsu_fossw')
        ml2_config.cfg.CONF.set_override(
            'password', 'password', 'fujitsu_fossw')
        ml2_config.cfg.CONF.set_override(
            'fossw_ips', [ADDRESS, ADDRESS2], 'fujitsu_fossw')
        importutils.import_object = mock.Mock()
        fossw_vlandriver.FOSSWVlanDriver().get_switch_mac_ip_pair = mock.Mock()

    def test_initialize_success(self):
        self.mech = mech_fossw.FOSSWMechanismDriver()
        self.assertEqual('admin', self.mech.username)
        self.assertEqual('password', self.mech.password)
        self.assertEqual('password', self.mech.password)
        self.assertEqual(
            mech_fossw.FOSSW_VLAN_DRIVER,
            importutils.import_object.call_args_list[0][0][0])
        self.assertEqual(
            ml2_config.cfg.CONF,
            importutils.import_object.call_args_list[0][0][1])
        self.assertEqual(
            mech_fossw.FOSSW_VXLAN_DRIVER,
            importutils.import_object.call_args_list[1][0][0])
        self.assertEqual(
            ml2_config.cfg.CONF,
            importutils.import_object.call_args_list[1][0][1])
        self.mech._vlan_driver.get_switch_mac_ip_pair.assert_called_once_with(
            [ADDRESS, ADDRESS2])

    def test_initialize_with_no_ips(self):
        ml2_config.cfg.CONF.set_override(
            'fossw_ips', [], 'fujitsu_fossw')
        self.assertRaises(
            cfg.RequiredOptError,
            mech_fossw.FOSSWMechanismDriver)


class TestFOSSWBaremetalPortsVlan(TestFujitsuMechDriverV2,
                                  helper.FujitsuMechanismHelper):

    def setUp(self):
        super(TestFOSSWBaremetalPortsVlan, self).setUp()

    def test_create_network(self):
        ctx = self.prepare_dummy_context('network')
        self.mech.create_network_postcommit(ctx)
        self.mech._vlan_driver.create_vlan.assert_called()
        self.mech._vxlan_driver.create_logical_switch.assert_not_called()
        self.assertEqual(2, self.mech._vlan_driver.create_vlan.call_count)

    def test_delete_network(self):
        ctx = self.prepare_dummy_context('network')
        self.mech.delete_network_postcommit(ctx)
        self.mech._vlan_driver.delete_vlan.assert_called()
        self.mech._vxlan_driver.delete_logical_switch.assert_not_called()
        self.assertEqual(2, self.mech._vlan_driver.delete_vlan.call_count)

    def test_create_network_raises_create_vlan(self):
        ctx = self.prepare_dummy_context('network')
        self.mech._vlan_driver.create_vlan.side_effect = Exception
        self.assertRaises(
            ml2_exc.MechanismDriverError,
            self.mech.create_network_postcommit, ctx
        )

    def test_delete_network_raises_delete_vlan(self):
        ctx = self.prepare_dummy_context('network')
        self.mech._vlan_driver.delete_vlan.side_effect = Exception
        self.assertRaises(
            ml2_exc.MechanismDriverError,
            self.mech.delete_network_postcommit, ctx
        )

    def test_create_port(self):
        ctx = self.prepare_dummy_context()
        self.mech.create_port_postcommit(ctx)
        self.mech._vxlan_driver.update_physical_port.assert_not_called()
        self.mech._vlan_driver.setup_vlan.assert_not_called()

    def test_bind_port_with_single(self):
        ctx = self.prepare_dummy_context()
        self.mech.bind_port(ctx)
        params = params_for_driver(ctx.current)
        self.mech._vlan_driver.setup_vlan.assert_called_once_with(
            params['segmentation_id'],
            params['local_link_info'],
            self.mech.switches_mac_ip_pair
        )

    def test_bind_port_with_single_raises_setup_vlan(self):
        ctx = self.prepare_dummy_context()
        self.mech._vlan_driver.setup_vlan.side_effect = Exception
        self.assertRaises(
            ml2_exc.MechanismDriverError,
            self.mech.bind_port, ctx)

    def test_bind_port_with_lag(self):
        ctx = self.prepare_dummy_context(nic='lag')
        self.mech.bind_port(ctx)
        params = params_for_driver(ctx.current, lag=True)
        self.mech._vlan_driver.setup_vlan_with_lag.assert_called_once_with(
            params['segmentation_id'],
            params['local_link_info'],
            self.mech.switches_mac_ip_pair
        )

    def test_bind_port_with_lag_raises_setup_vlan_with_lag(self):
        ctx = self.prepare_dummy_context(nic='lag')
        self.mech._vlan_driver.setup_vlan_with_lag.side_effect = Exception
        self.assertRaises(
            ml2_exc.MechanismDriverError,
            self.mech.bind_port, ctx)

    def test_bind_port_with_mlag(self):
        ctx = self.prepare_dummy_context(nic='mlag')
        self.mech.bind_port(ctx)
        params = params_for_driver(ctx.current, lag=True)
        self.mech._vlan_driver.setup_vlan_with_lag.assert_called_once_with(
            params['segmentation_id'],
            params['local_link_info'],
            self.mech.switches_mac_ip_pair
        )

    def test_bind_port_with_flat(self):
        ctx = self.prepare_dummy_context(net_type='flat')
        self.mech.bind_port(ctx)
        self.mech._vlan_driver.setup_vlan.assert_not_called()
        self.mech._vxlan_driver.update_physical_port.assert_not_called()

    def test_update_port(self):
        ctx = self.prepare_dummy_context(vif_type='other')
        self.mech.update_port_postcommit(ctx)
        self.mech._vlan_driver.clear_vlan.assert_not_called()

    def test_update_port_for_baremetal_with_unbound(self):
        ctx = self.prepare_dummy_context(vif_type='unbound', set_original=True)
        self.mech.update_port_postcommit(ctx)
        params = params_for_driver(ctx.original)
        self.mech._vlan_driver.clear_vlan.assert_called_with(
            params['local_link_info'],
            self.mech.switches_mac_ip_pair
        )

    def test_update_port_for_baremetal_with_unbound_and_lag(self):
        ctx = self.prepare_dummy_context(
            vif_type='unbound', set_original=True, nic='lag')
        self.mech.update_port_postcommit(ctx)
        params = params_for_driver(ctx.original)

        self.mech._vlan_driver.clear_vlan_with_lag.assert_called_once_with(
            params['local_link_info'],
            self.mech.switches_mac_ip_pair)

    def test_update_port_for_baremetal_with_unbound_flat(self):
        ctx = self.prepare_dummy_context(net_type='flat', vif_type='other',
                                         set_original=True)
        self.mech.update_port_postcommit(ctx)
        self.mech._vlan_driver.clear_vlan.assert_not_called()

    def test_update_port_with_bound_not_baremetal(self):
        ctx = self.prepare_dummy_context(vif_type='normal', vnic_type='ovs')
        self.mech.update_port_postcommit(ctx)
        self.mech._vlan_driver.associate_mac_to_network.assert_not_called()

    def test_delete_port_with_single(self):
        ctx = self.prepare_dummy_context(nic='single')
        self.mech.delete_port_postcommit(ctx)
        params = params_for_driver(ctx.current)
        self.mech._vlan_driver.clear_vlan.assert_called_with(
            params['local_link_info'],
            self.mech.switches_mac_ip_pair
        )

    def test_delete_with_lag(self):
        ctx = self.prepare_dummy_context(nic='lag')
        self.mech.delete_port_postcommit(ctx)
        params = params_for_driver(ctx.current, lag=True)
        self.mech._vlan_driver.clear_vlan_with_lag.assert_called_with(
            params['local_link_info'],
            self.mech.switches_mac_ip_pair
        )

    def test_delete_with_mlag(self):
        ctx = self.prepare_dummy_context(nic='mlag')
        self.mech.delete_port_postcommit(ctx)
        params = params_for_driver(ctx.current, lag=True)
        self.mech._vlan_driver.clear_vlan_with_lag.assert_called_with(
            params['local_link_info'],
            self.mech.switches_mac_ip_pair
        )

    def test_delete_port_with_normal(self):
        ctx = self.prepare_dummy_context(vnic_type='normal')
        self.mech.delete_port_postcommit(ctx)
        self.mech._vlan_driver.clear_vlan.assert_not_called()

    def test_delete_port_raises_driver_clear_vlan(self):
        self.mech._vlan_driver.clear_vlan.side_effect = Exception
        ctx = self.prepare_dummy_context(nic='single')
        self.assertRaises(
            ml2_exc.MechanismDriverError,
            self.mech.delete_port_postcommit, ctx)
        self.mech._vxlan_driver.reset_physical_port.assert_not_called()

    def test_delete_with_lag_raises_clear_vlan_with_lag(self):
        self.mech._vlan_driver.clear_vlan_with_lag.side_effect = Exception
        ctx = self.prepare_dummy_context(nic='lag')
        self.assertRaises(
            ml2_exc.MechanismDriverError,
            self.mech.delete_port_postcommit, ctx)
        self.mech._vxlan_driver.reset_physical_port.assert_not_called()


def params_for_driver(port, lag=False):
    return {
        'local_link_info': port['binding:profile']['local_link_information'],
        'segmentation_id': 1111,
        'mac': port['mac_address'],
        'lag': lag
    }


class TestFOSSWIsSupported(TestFujitsuMechDriverV2,
                           helper.FujitsuMechanismHelper):
    def setUp(self):
        super(TestFOSSWIsSupported, self).setUp()

    def test_is_supported_not_support_type(self):
        ctx = self.prepare_dummy_context('network', net_type='flat')
        network = ctx.current
        self.assertFalse(mech_fossw.is_supported(network))

    def test_is_supported_illegal_seg_id_is_none(self):
        ctx = self.prepare_dummy_context('network', net_type='vlan')
        network = ctx.current
        ctx.current['provider:segmentation_id'] = None
        self.assertFalse(mech_fossw.is_supported(network))

    def test_is_supported_illegal_seg_id_is_not_defined(self):
        ctx = self.prepare_dummy_context('network', net_type='vlan')
        network = ctx.current
        del network['provider:segmentation_id']
        self.assertFalse(mech_fossw.is_supported(network))


class TestFOSSWBaremetalPortsVxlan(TestFujitsuMechDriverV2,
                                   helper.FujitsuMechanismHelper):

    def setUp(self):
        super(TestFOSSWBaremetalPortsVxlan, self).setUp()

    def test_is_supported_not_support_type(self):
        ctx = self.prepare_dummy_context('network', net_type='flat')
        network = ctx.current
        self.assertFalse(mech_fossw.is_supported(network))

    def test_is_supported_illegal_seg_id_is_none(self):
        ctx = self.prepare_dummy_context('network', net_type='flat')
        network = ctx.current
        ctx.current['provider:segmentation_id'] = None
        self.assertFalse(mech_fossw.is_supported(network))

    def test_is_supported_illegal_seg_id_is_not_defined(self):
        ctx = self.prepare_dummy_context('network', net_type='flat')
        network = ctx.current
        self.assertFalse(mech_fossw.is_supported(network))

    def test_create_network(self):
        ctx = self.prepare_dummy_context('network', net_type='vxlan')
        self.mech.create_network_postcommit(ctx)
        self.mech._vlan_driver.create_vlan.assert_not_called()
        net_id = ctx.current['id']
        vni = ctx.current['provider:segmentation_id']
        self.mech._vxlan_driver.create_logical_switch.assert_called_with(
            net_id, vni)

    def test_delete_network(self):
        ctx = self.prepare_dummy_context('network', net_type='vxlan')
        self.mech.delete_network_postcommit(ctx)
        net_id = ctx.current['id']
        self.mech._vlan_driver.delete_vlan.assert_not_called()
        self.mech._vxlan_driver.delete_logical_switch.assert_called_with(
            net_id)

    def test_create_network_raises_create_logical_switch(self):
        ctx = self.prepare_dummy_context('network', net_type='vxlan')
        self.mech._vxlan_driver.create_logical_switch.side_effect = Exception
        self.assertRaises(
            ml2_exc.MechanismDriverError,
            self.mech.create_network_postcommit, ctx
        )

    def test_delete_network_raises_delete_logical_switch(self):
        ctx = self.prepare_dummy_context('network', net_type='vxlan')
        self.mech._vxlan_driver.delete_logical_switch.side_effect = Exception
        self.assertRaises(
            ml2_exc.MechanismDriverError,
            self.mech.delete_network_postcommit, ctx
        )

    def test_create_port(self):
        ctx = self.prepare_dummy_context(net_type='vxlan')
        self.mech.create_port_postcommit(ctx)
        self.mech._vlan_driver.setup_vlan.assert_not_called()
        self.mech._vxlan_driver.update_physical_port.assert_not_called()

    def test_bind_port_with_single(self):
        ctx = self.prepare_dummy_context(net_type='vxlan')
        self.mech.bind_port(ctx)
        params = params_for_driver(ctx.current)
        self.mech._vxlan_driver.update_physical_port.assert_called_once_with(
            params['segmentation_id'],
            params['local_link_info'],
            ctx.current,
            self.mech.switches_mac_ip_pair,
            ctx.network._plugin_context.request_id
        )

    def test_bind_port_with_lag(self):
        ctx = self.prepare_dummy_context(net_type='vxlan', nic='lag')
        self.mech.bind_port(ctx)
        params = params_for_driver(ctx.current, lag=True)
        self.mech._vlan_driver.setup_vlan_with_lag.assert_called_once_with(
            mech_fossw.DEFAULT_VLAN,
            params['local_link_info'],
            self.mech.switches_mac_ip_pair)
        target = self.mech._vxlan_driver.update_physical_port_with_lag
        target.assert_called_once_with(
            params['segmentation_id'],
            params['local_link_info'],
            ctx.current,
            self.mech.switches_mac_ip_pair,
            ctx.network._plugin_context.request_id
        )

    def test_bind_port_with_mlag(self):
        ctx = self.prepare_dummy_context(net_type='vxlan', nic='mlag')
        self.mech.bind_port(ctx)
        params = params_for_driver(ctx.current, lag=True)
        self.mech._vlan_driver.setup_vlan_with_lag.assert_called_once_with(
            mech_fossw.DEFAULT_VLAN,
            params['local_link_info'],
            self.mech.switches_mac_ip_pair)
        target = self.mech._vxlan_driver.update_physical_port_with_lag
        target.assert_called_once_with(
            params['segmentation_id'],
            params['local_link_info'],
            ctx.current,
            self.mech.switches_mac_ip_pair,
            ctx.network._plugin_context.request_id
        )

    def test_bind_port_with_lag_raises_driver_method(self):
        ctx = self.prepare_dummy_context(net_type='vxlan', nic='lag')
        self.mech._vxlan_driver.update_physical_port_with_lag.side_effect = \
            Exception
        self.assertRaises(
            ml2_exc.MechanismDriverError,
            self.mech.bind_port, ctx)

    def test_update_port(self):
        ctx = self.prepare_dummy_context(net_type='vxlan', vif_type='other')
        self.mech.update_port_postcommit(ctx)
        self.mech._vxlan_driver.reset_physical_port.assert_not_called()

    def test_update_port_with_flat(self):
        ctx = self.prepare_dummy_context(net_type='flat', vif_type='other')
        self.mech.update_port_postcommit(ctx)
        self.mech._vxlan_driver.update_physical_port.assert_not_called()

    def test_update_port_with_bound(self):
        ctx = self.prepare_dummy_context(net_type='vxlan', vif_type='normal')
        self.mech.update_port_postcommit(ctx)
        self.mech._vxlan_driver.update_physical_port.assert_not_called()

    def test_update_port_for_baremetal_with_unbound(self):
        ctx = self.prepare_dummy_context(net_type='vxlan', vif_type='unbound',
                                         set_original=True)
        self.mech.update_port_postcommit(ctx)
        params = params_for_driver(ctx.original)
        self.mech._vxlan_driver.reset_physical_port.assert_called_with(
            params['local_link_info'],
            ctx.original,
            self.mech.switches_mac_ip_pair
        )

    def test_update_port_for_baremetal_with_unbound_and_lag(self):
        ctx = self.prepare_dummy_context(
            net_type='vxlan', vif_type='unbound', set_original=True, nic='lag')
        self.mech.update_port_postcommit(ctx)
        params = params_for_driver(ctx.original)

        self.mech._vlan_driver.clear_vlan_with_lag.assert_called_once_with(
            params['local_link_info'],
            self.mech.switches_mac_ip_pair)
        target = self.mech._vxlan_driver.reset_physical_port_with_lag
        target.assert_called_with(
            params['local_link_info'],
            ctx.original,
            self.mech.switches_mac_ip_pair
        )

    def test_update_port_for_baremetal_with_unbound_flat(self):
        ctx = self.prepare_dummy_context(net_type='flat', vif_type='other',
                                         set_original=True)
        self.mech.clear_vxlan = mock.Mock()
        self.mech.update_port_postcommit(ctx)
        self.mech.clear_vxlan.assert_not_called()

    def test_update_port_with_bound_not_baremetal(self):
        ctx = self.prepare_dummy_context(
            net_type='vxlan', vif_type='normal', vnic_type='ovs')
        self.mech.update_port_postcommit(ctx)
        params = params_for_driver(ctx.current)
        self.mech._vxlan_driver.update_physical_port.assert_called_with(
            params['segmentation_id'],
            [],
            ctx.current,
            self.mech.switches_mac_ip_pair
        )

    def test_update_port_raises_update_physical_port(self):
        ctx = self.prepare_dummy_context(
            net_type='vxlan', vif_type='normal', vnic_type='ovs')
        self.mech._vxlan_driver.update_physical_port.side_effect = Exception
        self.assertRaises(
            ml2_exc.MechanismDriverError,
            self.mech.update_port_postcommit, ctx)

    def test_delete_port(self):
        ctx = self.prepare_dummy_context(net_type='vxlan')
        self.mech.delete_port_postcommit(ctx)
        self.mech._vxlan_driver.reset_physical_port.assert_not_called()

    def test_delete_port_with_bound(self):
        ctx = self.prepare_dummy_context(net_type='vxlan', vif_type='normal')
        self.mech.delete_port_postcommit(ctx)
        params = params_for_driver(ctx.current)
        self.mech._vxlan_driver.reset_physical_port.assert_called_with(
            params['local_link_info'],
            ctx.current,
            self.mech.switches_mac_ip_pair
        )

    def test_delete_port_raises_reset_physical_port(self):
        ctx = self.prepare_dummy_context(net_type='vxlan', vif_type='normal')
        self.mech._vxlan_driver.reset_physical_port.side_effect = Exception
        self.assertRaises(
            ml2_exc.MechanismDriverError,
            self.mech.delete_port_postcommit, ctx)

    def test_delete_port_with_lag_raises_reset_physical_port_with_lag(self):
        ctx = self.prepare_dummy_context(
            net_type='vxlan', vif_type='normal', nic='lag')
        self.mech._vlan_driver.clear_vlan_with_lag.side_effect = [
            Exception, None]
        self.mech._vxlan_driver.reset_physical_port_with_lag.side_effect = [
            Exception, Exception]
        self.assertRaises(
            ml2_exc.MechanismDriverError,
            self.mech.delete_port_postcommit, ctx)
        self.assertRaises(
            ml2_exc.MechanismDriverError,
            self.mech.delete_port_postcommit, ctx)

    def test_delete_port_with_flat(self):
        ctx = self.prepare_dummy_context(net_type='flat')
        self.mech.delete_port_postcommit(ctx)
        self.mech._vxlan_driver.reset_physical_port.assert_not_called()

    def test_delete_with_lag(self):
        ctx = self.prepare_dummy_context(
            net_type='vxlan', vif_type='normal', nic='lag')
        self.mech.delete_port_postcommit(ctx)
        params = params_for_driver(ctx.current)
        target = self.mech._vxlan_driver.reset_physical_port_with_lag
        target.assert_called_with(
            params['local_link_info'],
            ctx.current,
            self.mech.switches_mac_ip_pair
        )

    def test_delete_with_mlag(self):
        ctx = self.prepare_dummy_context(
            net_type='vxlan', vif_type='normal', nic='mlag')
        self.mech.delete_port_postcommit(ctx)
        params = params_for_driver(ctx.current)
        target = self.mech._vxlan_driver.reset_physical_port_with_lag
        target.assert_called_with(
            params['local_link_info'],
            ctx.current,
            self.mech.switches_mac_ip_pair
        )



# Copyright 2017 FUJITSU LIMITED
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from neutron.db.models.plugins.ml2 import vxlanallocation as vxlan_model
from neutron.plugins.ml2.drivers import type_tunnel


class TypeVxlan(type_tunnel.EndpointTunnelTypeDriver):
    def __init__(self):
        super(TypeVxlan, self).__init__(vxlan_model.VxlanAllocation,
                                        vxlan_model.VxlanEndpoints)

    def get_type(self):
        return "vxlan"

    def initialize(self):
        pass

    def get_endpoints(self):
        """Get every vxlan endpoints from Neutron DB."""
        vxlan_endpoints = self._get_endpoints()
        return [{'ip_address': vxlan_endpoint.ip_address,
                 'udp_port': vxlan_endpoint.udp_port,
                 'host': vxlan_endpoint.host}
                for vxlan_endpoint in vxlan_endpoints]

    def add_endpoint(self, ip, host, udp_port):
        """Add switch endpoint information to Neutron DB."""
        return self._add_endpoint(ip, host, udp_port=udp_port)

    def db_get_endpoint_by_ip(self, ip):
        """Get vxlan endpoint with given IPaddress from DB."""
        vxlan_endpoint = self.get_endpoint_by_ip(ip)
        return {'ip_address': vxlan_endpoint.ip_address,
                'udp_port': vxlan_endpoint.udp_port,
                'host': vxlan_endpoint.host}

    def db_delete_endpoint(self, ip):
        """Delete switch endpoint information from DB."""
        self.delete_endpoint(ip)

    def db_get_endpoint_ip_by_host(self, host):
        """Get vxlan endpoint IP address with given host name from DB."""
        vxlan_endpoint = self.get_endpoint_by_host(host)
        return vxlan_endpoint.ip_address



# -*- coding: utf-8 -*-

# Copyright 2015-2016 FUJITSU LIMITED
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_networking_fujitsu
----------------------------------

Tests for `networking_fujitsu` module.
"""
from networking_fujitsu.ml2.ism import mech_ism
from neutron.tests.unit.plugins.ml2 import test_plugin as test_ml2_plugin


class FujitsuIsmDriverTestCase(test_ml2_plugin.Ml2PluginV2TestCase):

    _mechanism_drivers = ['fujitsu_ism']
    # test_create_router_port_and_fail_create_postcommit:
    #     This one is mocked 'fake_driver' only. So, our plugin's instance
    #     hasn't mocked and fail definitely. Therefore, skips this test.
    _skip = ["test_create_router_port_and_fail_create_postcommit"]

    def setUp(self):

        if self._testMethodName in self._skip:
            self.skipTest("This test has already verified at neutron's test.")
        self._driver = mech_ism.ISMMechanismDriver()
        super(FujitsuIsmDriverTestCase, self).setUp()

    def tearDown(self):
        super(FujitsuIsmDriverTestCase, self).tearDown()



# Copyright 2017 FUJITSU LIMITED
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Fujitsu fossw Driver implements CLI over ssh for Neutron network life-cycle
management.
"""

import copy

from oslo_config import cfg
from oslo_log import log as logging

from networking_fujitsu._i18n import _LE
from networking_fujitsu._i18n import _LW
from networking_fujitsu.ml2.fossw import client
from neutron.common import utils

LOG = logging.getLogger(__name__)
_LOCK_NAME = 'fujitsu-fossw'


class FOSSWVlanDriver(object):
    """Fossw CLI interface driver for Neutron network.

    Handles life-cycle management of Neutron network.
    """

    def __init__(self, conf=None):
        if conf:
            self._conf = conf
        else:
            self._conf = cfg.CONF
        self.client = client.FOSSWClient(self._conf)

    @utils.synchronized(_LOCK_NAME)
    def get_switch_mac_ip_pair(self, ips):
        """Get and create pair of switch's MAC address and IP address.

        :param ips: list of IP address of FOS switch
        :type ips: list of IP address strings

        :returns: the dictionary which stores pair of  MAC address and IP
                  address
        :rtype: dictionary of MAC address and IP address

        """
        switches_mac_ip_pair = {}
        for ip in ips:
            self.client.connect(ip)
            mac = self.client.get_switch_mac()
            if not mac:
                self.client.disconnect()
                LOG.exception(
                    _LE('Cannot get MAC address from FOS switch(%s)'), ip)
                raise client.FOSSWClientException('get_switch_mac_ip_pair')
            switches_mac_ip_pair.update({mac: ip})
            self.client.disconnect()
        return switches_mac_ip_pair

    def create_vlan(self, ip, vlanid):
        """Create VLAN with specified VLAN ID.

        :param ip: IP address of target FOS switch
        :type ip: string
        :param vlanid: ID of VLAN to be created
        :type vlanid: string

        :returns: None
        :rtype: None

        """
        method = 'create_vlan'
        try:
            self.client.connect(ip)
            self.client.create_vlan(vlanid)
            self.client.disconnect()
        except Exception as e:
            self.client.disconnect()
            LOG.exception(_LE("an error occurred while creating vlan to FOS "
                              "switch. %s"), e)
            raise client.FOSSWClientException(method)

    def delete_vlan(self, ip, vlanid):
        """Delete VLAN with specified VLAN ID.

        :param ip: IP address of target FOS switch
        :type ip: string
        :param vlanid: ID of VLAN to be deleted
        :type vlanid: string

        :returns: None
        :rtype: None

        """
        method = 'delete_vlan'
        try:
            self.client.connect(ip)
            self.client.delete_vlan(vlanid)
            self.client.disconnect()
        except Exception as e:
            self.client.disconnect()
            LOG.exception(_LE("An error occurred while deleting vlan from "
                              "FOS switch. %s"), e)
            raise client.FOSSWClientException(method)

    def setup_vlan(self, vlanid, lli, ip_mac_pairs):
        """Setup VLAN for physical port on FOS Switch.

        :param vlanid: the ID of VLAN to be associated
        :type vlanid: string
        :param lli: the local link information of ironic node
        :type lli: list of local link information dictionary
        :param ip_mac_pairs: the pair of MAC address and IP address of FOS
                             switch
        :type ip_mac_pairs: dictionary

        :returns: None
        :rtype: None

        """
        method = "setup_vlan"
        sw_mac = lli[0]['switch_id']
        port_id = lli[0]['port_id']

        try:
            target_ip = ip_mac_pairs[sw_mac]
        except KeyError as e:
            LOG.exception(_LE("FOS switch with MAC(%s) not found."), sw_mac)
            raise client.FOSSWClientException(method)
        try:
            self.client.connect(target_ip)
            self.client.set_vlan(vlanid, port_id)
            self.client.disconnect()
        except Exception as e:
            self.client.disconnect()
            LOG.exception(_LE("An error occurred while setup vlan for "
                              "physical port on FOS switch. %s"), e)
            raise client.FOSSWClientException(method)

    @utils.synchronized(_LOCK_NAME, external=True)
    def setup_vlan_with_lag(self, vlanid, llis, ip_mac_pairs):
        """Setup VLAN and LAG for physical ports FOS switch.

        :param vlanid: the ID of VLAN to be associated
        :type vlanid: string
        :param llis: the local link informations of ironic node
        :type llis: list of local link informations dictionary
        :param ip_mac_pairs: the pair of MAC address and IP address of FOS
                             switch
        :type ip_mac_pairs: dictionary

        :returns: None
        :rtype: None

        """
        method = "setup_vlan_with_lag"
        mlag = False
        unique_mac_list = sorted(list(set([lli['switch_id'] for lli in llis])))
        if len(unique_mac_list) > 1:
            mlag = self.is_valid_mlag(unique_mac_list, ip_mac_pairs)
            if not mlag:
                LOG.exception(
                    _LE("Specified switches cannot set mLAG pair. "
                        "Please confirm each switch's peerlink setting.")
                )
                raise client.FOSSWClientException(method)

        # NOTE(takanorimiyagishi): Currently this driver is hard-coded for
        # LAG(802.3ad) setting for FOS switch. When logical ports on FOS switch
        # are set as LAG interface, FOS switch skips VLAN setting of physical
        # port.
        #
        # Setup VLAN for each physical port
        # for lli in llis:
        #     self.setup_vlan(vlanid, [lli], ip_mac_pairs)

        # Create lag resource
        for mac in unique_mac_list:
            target_ip = ip_mac_pairs[mac]
            mac_llis = [lli for lli in llis if lli['switch_id'] == mac]
            ports = [m_lli['port_id'] for m_lli in mac_llis]
            self.client.connect(target_ip)
            lag_port = self.client.get_lag_port()
            if not lag_port:
                self.client.disconnect()
                LOG.exception(_LE("Could not find available logicalport in "
                                  "switch(%s)."), target_ip)
                raise client.FOSSWClientException(method)

            for port in ports:
                self.client.join_to_lag(port, lag_port)

            # Setup VLAN for logical port
            lag_lli = copy.deepcopy(mac_llis[0])
            lag_lli['port_id'] = lag_port
            self.setup_vlan(vlanid, [lag_lli], ip_mac_pairs)

            if mlag:
                # Get available VPC id from FOS switch
                vpcid = self.client.get_vpcid()
                if vpcid:
                    self.client.join_to_vpc(lag_port, vpcid)
                else:
                    # All vpc is already used by other logical ports.
                    self.client.disconnect()
                    raise client.FOSSWClientException(method)
            self.client.disconnect()

    def _validate_lli_macs_with_config(self, macs, ip_mac_pairs):
        ips = [ip_mac_pairs.get(mac, None) for mac in macs]
        if None in ips:
            LOG.error(_LE("MAC(%s) in local_link_informatio doesn't match "
                          "with FOS switches"), macs)
            raise client.FOSSWClientException('_validate_lli_macs_with_config')
        return ips

    def is_valid_mlag(self, macs, ip_mac_pairs):
        """Validates which given pair of MAC address is valid mlag pair or not.

        :param macs: list of MAC address which mlag candidacy pair
        :type macs: list
        :param ip_mac_pairs: the pair of MAC address and IP address of FOS
                             switches
        :type ip_mac_pairs: dictionary

        :returns: Result of validation
        :rtype: Boolean

        """
        try:
            ips = self._validate_lli_macs_with_config(macs, ip_mac_pairs)
            self.client.connect(ips[0])
            partner_ip = self.client.get_peerlink_partner()
            self.client.disconnect()
            return (partner_ip in ips)
        except Exception as e:
            self.client.disconnect()
            LOG.warning(_LW("An error occurred while validating specified "
                            "FOS switches are VPC pair. %s"), e)
            return False

    def clear_vlan(self, lli, ip_mac_pairs):
        """Clear VLAN from FOS switch.

        :param lli: the local link information of ironic node
        :type lli: list
        :param ip_mac_pairs: the pair of MAC address and IP address of FOS
                             switches
        :type ip_mac_pairs: dictionary
        """
        method = "clear_vlan"
        sw_mac = lli[0]['switch_id']
        port_id = lli[0]['port_id']

        try:
            target_ip = ip_mac_pairs[sw_mac]
        except KeyError as e:
            LOG.exception(_LE("FOS switch with MAC(%s) not found."), sw_mac)
            raise client.FOSSWClientException(method)
        try:
            self.client.connect(target_ip)
            self.client.clear_vlan(port_id)
            self.client.disconnect()
        except Exception as e:
            self.client.disconnect()
            LOG.exception(_LE("an error occurred while clearing VLAN from "
                              "physical port on FOS switch. %s"), e)
            raise client.FOSSWClientException(method)

    @utils.synchronized(_LOCK_NAME, external=True)
    def clear_vlan_with_lag(self, llis, ip_mac_pairs):
        """Clear VLAN and LAG from FOS switch.

        :param llis: the local link informations of ironic node
        :type llis: list
        :param ip_mac_pairs: the pair of MAC address and IP address of FOS
                             switches
        :type ip_mac_pairs: dictionary

        :returns: None
        :rtype: None

        """
        method = "clear_vlan_with_lag"
        mlag = None
        unique_mac_list = sorted(list(set([lli['switch_id'] for lli in llis])))
        if len(unique_mac_list) > 1:
            mlag = self.is_valid_mlag(unique_mac_list, ip_mac_pairs)
            if not mlag:
                LOG.exception(_LE("Specified switches are not able to clear "
                                  "mLAG pair. Please confirm each switch's "
                                  "peerlink setting."))
                raise client.FOSSWClientException(method)

        for mac in unique_mac_list:
            target_ip = ip_mac_pairs[mac]
            mac_llis = [lli for lli in llis if lli['switch_id'] == mac]
            ports = [m_lli['port_id'] for m_lli in mac_llis]
            self.client.connect(target_ip)
            lag_port = self.client.get_lag_port(mac_llis[0]['port_id'])
            if lag_port:
                if mlag:
                    vpcid = self.client.get_vpcid(lag_port)
                    if vpcid:
                        self.client.leave_from_vpc(lag_port, vpcid)
                    else:
                        LOG.warning(
                            _LW("Specified logicalport has been already "
                                "disssociated with any VPC on the switch(%s). "
                                "skip leave_from_vpc."), target_ip)
                for port in ports:
                    self.client.leave_from_lag(port, lag_port)

                # Clear VLAN for logical port
                lag_lli = copy.deepcopy(mac_llis[0])
                lag_lli['port_id'] = lag_port
                self.clear_vlan([lag_lli], ip_mac_pairs)
            else:
                LOG.warning(
                    _LW("Specified logicalport has already cleared. Skip "
                        "clearing LAG."))
            self.client.disconnect()
        # NOTE(takanorimiyagishi): Currently this driver is hard-coded for
        # LAG(802.3ad) setting for FOS switch. When logical port on FOS switch
        # are set as LAG interface, FOS switch ignores VLAN setting of physical
        # port which belongs to the LAG interface.
        #
        # Clear VLAN for each physical port
        # for lli in llis:
        #     self.clear_vlan([lli], ip_mac_pairs)



# Copyright 2017 FUJITSU LIMITED
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import copy
import mock
from networking_fujitsu.ml2.fossw import client
from networking_fujitsu.ml2.fossw import fossw_vlandriver
from networking_fujitsu.ml2.fossw import mech_fossw
from neutron.tests import base
from oslo_config import cfg

LLI = {'single': [{"switch_id": "00:00:4c:ee:e5:39", "port_id": "0/1",
                   "switch_info": "ET-7648BRA-FOS"}],
       'lag': [{"switch_id": "00:00:4c:ee:e5:39", "port_id": "0/1",
                "switch_info": "ET-7648BRA-FOS"},
               {"switch_id": "00:00:4c:ee:e5:39", "port_id": "0/2",
                "switch_info": "ET-7648BRA-FOS"}],
       'mlag': [{"switch_id": "00:00:4c:ee:e5:39", "port_id": "0/1",
                 "switch_info": "ET-7648BRA-FOS"},
                {"switch_id": "00:00:4c:ee:e5:40", "port_id": "0/1",
                 "switch_info": "ET-7648BRA-FOS"}]}
DUMMY_IP = '192.168.0.1'
DUMMY_IPS = ['192.168.0.1', '192.168.0.2']


class BaseTestFOSSWVlanDriver(base.BaseTestCase):
    """Base class to test Fujitsu FOSSW driver."""

    def setUp(self):
        super(BaseTestFOSSWVlanDriver, self).setUp()
        cfg.CONF.register_opts(
            mech_fossw.ML2_FUJITSU,
            mech_fossw.ML2_FUJITSU_GROUP
        )
        self.driver = fossw_vlandriver.FOSSWVlanDriver(cfg.CONF)


class TestFOSSWVlanDriverGetSwitchMacIpPair(BaseTestFOSSWVlanDriver):
    """Test Fujitsu FOSSW vlan driver for get_switch_mac_ip_pair"""
    def setUp(self):
        super(TestFOSSWVlanDriverGetSwitchMacIpPair, self).setUp()
        self.driver.client = mock.Mock()
        self.driver.client.get_switch_mac.return_value = 'aa:aa:aa:aa:aa:aa'

    def test_returns_mac_address_with_single_ip(self):
        ret = self.driver.get_switch_mac_ip_pair([DUMMY_IP])
        self.assertEqual({'aa:aa:aa:aa:aa:aa': DUMMY_IP}, ret)

    def test_returns_mac_address_with_multiple_ips(self):
        self.driver.client.get_switch_mac.side_effect = [
            'aa:aa:aa:aa:aa:aa',
            'ff:ff:ff:ff:ff:ff']
        expect = {
            'aa:aa:aa:aa:aa:aa': '192.168.0.1',
            'ff:ff:ff:ff:ff:ff': '192.168.0.2'}
        ret = self.driver.get_switch_mac_ip_pair(DUMMY_IPS)
        self.assertEqual(expect, ret)

    def test_mac_cannot_get(self):
        self.driver.client.get_switch_mac.return_value = None
        self.assertRaises(
            client.FOSSWClientException,
            self.driver.get_switch_mac_ip_pair, [DUMMY_IP])

    def test_mac_cannot_get_with_multiple_ips(self):
        self.driver.client.get_switch_mac.return_value = None
        self.assertRaises(
            client.FOSSWClientException,
            self.driver.get_switch_mac_ip_pair, DUMMY_IPS)


class TestFOSSWVlanDriverCreateVlan(BaseTestFOSSWVlanDriver):
    """Test Fujitsu fossw driver for create VLAN"""

    def setUp(self):
        super(TestFOSSWVlanDriverCreateVlan, self).setUp()
        self.driver.client.connect = mock.Mock(return_value=None)
        self.driver.client.create_vlan = mock.Mock(return_value=None)
        self.driver.client.disconnect = mock.Mock(return_value=None)

    def test_create_vlan(self):
        self.driver.create_vlan(DUMMY_IP, 2)
        self.assertIsNone(self.driver.client.ssh)

    def test_create_vlan_fail(self):
        error = client.FOSSWClientException('create_vlan')
        self.driver.client.connect.side_effect = error
        self.assertRaises(
            client.FOSSWClientException,
            self.driver.create_vlan,
            DUMMY_IP,
            2
        )


class TestFOSSWVlanDriverDeleteVlan(BaseTestFOSSWVlanDriver):
    """Test Fujitsu FOSSW vlan driver for delete VLAN"""
    def setUp(self):
        super(TestFOSSWVlanDriverDeleteVlan, self).setUp()
        self.driver.client.connect = mock.Mock(return_value=None)
        self.driver.client.delete_vlan = mock.Mock(return_value=None)
        self.driver.client.disconnect = mock.Mock(return_value=None)

    def test_delete_vlan(self):
        self.driver.delete_vlan(DUMMY_IP, 2)
        self.assertIsNone(self.driver.client.ssh)

    def test_delete_vlan_fail(self):
        error = client.FOSSWClientException('delete_vlan')
        self.driver.client.connect.side_effect = error
        self.assertRaises(
            client.FOSSWClientException,
            self.driver.delete_vlan,
            DUMMY_IP,
            2
        )


class TestFOSSWVlanDriverSetupVlan(BaseTestFOSSWVlanDriver):
    """Test Fujitsu FOSSW vlan driver."""
    def setUp(self):
        super(TestFOSSWVlanDriverSetupVlan, self).setUp()
        self.driver.client.connect = mock.Mock(return_value=None)
        self.driver.client.set_vlan = mock.Mock(return_value=None)
        self.driver.client.disconnect = mock.Mock(return_value=None)

    def test_setup_vlan(self):
        ip_mac_pairs = {'00:00:4c:ee:e5:39': '192.168.1.1'}
        self.driver.setup_vlan(2, LLI['single'], ip_mac_pairs)
        self.assertIsNone(self.driver.client.ssh)

    def test_setup_vlan_fail_with_keyerror(self):
        ip_mac_pairs = {'00:00:4c:ee:e5:40': '192.168.1.1'}
        self.assertRaises(
            client.FOSSWClientException,
            self.driver.setup_vlan,
            2,
            LLI['single'],
            ip_mac_pairs
        )
        self.driver.client.disconnect.assert_not_called()

    def test_setup_vlan_fail_with_connect_fail(self):
        ip_mac_pairs = {'00:00:4c:ee:e5:39': '192.168.1.1'}
        error = client.FOSSWClientException('set_vlan')
        self.driver.client.connect.side_effect = error
        self.assertRaises(
            client.FOSSWClientException,
            self.driver.setup_vlan,
            2,
            LLI['single'],
            ip_mac_pairs
        )
        self.driver.client.set_vlan.assert_not_called()
        self.driver.client.disconnect.assert_called_once()


class TestFOSSWVlanDriverIsValidMLAG(BaseTestFOSSWVlanDriver):
    """Test FujitsuFOSSW vlan driver for setup vlan with lag"""
    def setUp(self):
        super(TestFOSSWVlanDriverIsValidMLAG, self).setUp()
        self.driver.client = mock.Mock()

    def test_valid(self):
        ip_mac_pairs = {"00:00:4c:ee:e5:39": "192.168.1.1",
                        "00:00:4c:ee:e5:40": "192.168.1.2"}
        macs = ["00:00:4c:ee:e5:39", "00:00:4c:ee:e5:40"]
        self.driver.client.get_peerlink_partner.return_value = '192.168.1.1'
        self.assertTrue(self.driver.is_valid_mlag(macs, ip_mac_pairs))

    def test_invalid_with_non_exist_mac(self):
        ip_mac_pairs = {"00:00:4c:ee:e5:39": "192.168.1.1",
                        "00:00:4c:ee:e5:40": "192.168.1.2"}
        macs = ["00:00:4c:ee:e5:39", "ff:ff:ff:ff:ff:ff"]
        self.assertFalse(self.driver.is_valid_mlag(macs, ip_mac_pairs))

    def test_raises_with_get_peerlink_partner(self):
        ip_mac_pairs = {"00:00:4c:ee:e5:39": "192.168.1.1",
                        "00:00:4c:ee:e5:40": "192.168.1.2"}
        macs = ["00:00:4c:ee:e5:39", "ff:ff:ff:ff:ff:ff"]
        self.driver.client.get_peerlink_partner.side_effect = Exception
        self.assertFalse(self.driver.is_valid_mlag(macs, ip_mac_pairs))


class TestFOSSWVlanDriverSetupVlanWithLAG(BaseTestFOSSWVlanDriver):
    """Test FujitsuFOSSW vlan driver for setup vlan with lag"""
    def setUp(self):
        super(TestFOSSWVlanDriverSetupVlanWithLAG, self).setUp()
        self.driver.is_valid_mlag = mock.Mock(return_value=False)
        self.driver.setup_vlan = mock.Mock()
        self.driver.client = mock.Mock()
        self.driver.client.get_lag_port.return_value = "3/1"
        self.driver.client.get_vpcid.return_value = 1
        self.vlanid = 2

    def test_lag_normal(self):
        llis = LLI['lag']
        lag_lli = copy.deepcopy(llis[0])
        lag_lli['port_id'] = '3/1'
        ip_mac_pairs = {"00:00:4c:ee:e5:39": "192.168.1.1",
                        "00:00:4c:ee:e5:40": "192.168.1.2"}
        self.assertIsNone(self.driver.setup_vlan_with_lag(
            self.vlanid, llis, ip_mac_pairs))

        self.driver.is_valid_mlag.assert_not_called()
        self.driver.client.connect.assert_called_once_with(
            ip_mac_pairs[llis[0]['switch_id']])
        self.driver.client.get_lag_port.assert_called_once()
        self.assertEqual(2, self.driver.client.join_to_lag.call_count)
        for lli in llis:
            self.driver.client.join_to_lag.assert_called_with(mock.ANY, '3/1')
        self.driver.setup_vlan.assert_called_once_with(
            self.vlanid, [lag_lli], ip_mac_pairs)
        self.driver.client.get_vpcid.assert_not_called()
        self.driver.client.join_to_vpc.assert_not_called()
        self.driver.client.disconnect.assert_called_once()

    def test_lag_lagport_not_found(self):
        self.driver.client.get_lag_port.return_value = None
        llis = LLI['lag']
        ip_mac_pairs = {"00:00:4c:ee:e5:39": "192.168.1.1",
                        "00:00:4c:ee:e5:40": "192.168.1.2"}
        self.assertRaises(
            client.FOSSWClientException,
            self.driver.setup_vlan_with_lag,
            self.vlanid, llis, ip_mac_pairs)
        self.driver.client.disconnect.assert_called_once()
        self.driver.client.join_to_vpc.assert_not_called()
        self.driver.client.get_vpcid.assert_not_called()
        self.driver.client.join_to_lag.assert_not_called()
        self.driver.setup_vlan.assert_not_called()

    def test_mlag_normal(self):
        llis = LLI['mlag']
        ip_mac_pairs = {"00:00:4c:ee:e5:39": "192.168.1.1",
                        "00:00:4c:ee:e5:40": "192.168.1.2"}
        self.driver.is_valid_mlag.return_value = True
        self.driver.client.join_to_vpc = mock.Mock(return_value=None)
        self.assertIsNone(self.driver.setup_vlan_with_lag(
            self.vlanid, llis, ip_mac_pairs))
        unique_mac_list = sorted(["00:00:4c:ee:e5:39", "00:00:4c:ee:e5:40"])
        self.driver.is_valid_mlag.assert_called_once()

        self.assertEqual(
            unique_mac_list, self.driver.is_valid_mlag.call_args[0][0])
        self.assertEqual(
            ip_mac_pairs, self.driver.is_valid_mlag.call_args[0][1])
        self.assertEqual(2, self.driver.client.join_to_vpc.call_count)

        lag_p = [{
            'switch_id': unique_mac_list[0],
            'port_id': '3/1',
            'switch_info': 'ET-7648BRA-FOS'}, {
            'switch_id': unique_mac_list[1],
            'port_id': '3/1',
            'switch_info': 'ET-7648BRA-FOS',
        }]
        for idx, arg in enumerate(self.driver.setup_vlan.call_args_list):
            self.assertEqual(self.vlanid, arg[0][0])
            self.assertEqual([lag_p[idx]], arg[0][1])
            self.assertEqual(ip_mac_pairs, arg[0][2])

        for arg in self.driver.client.join_to_vpc.call_args_list:
            self.assertEqual(('3/1', 1), arg[0])

    def test_mlag_not_valid_pair(self):
        llis = LLI['mlag']
        ip_mac_pairs = {"00:00:4c:ee:e5:39": "192.168.1.1",
                        "00:00:4c:ee:e5:40": "192.168.1.2"}
        self.assertRaises(client.FOSSWClientException,
                          self.driver.setup_vlan_with_lag,
                          self.vlanid,
                          llis,
                          ip_mac_pairs)
        self.driver.client.setup_vlan.assert_not_called()

    def test_mlag_vpcid_already_used(self):
        llis = LLI['mlag']
        ip_mac_pairs = {"00:00:4c:ee:e5:39": "192.168.1.1",
                        "00:00:4c:ee:e5:40": "192.168.1.2"}
        self.driver.is_valid_mlag.return_value = True
        self.driver.client.get_vpcid = mock.Mock(return_value=None)
        self.assertRaises(client.FOSSWClientException,
                          self.driver.setup_vlan_with_lag,
                          self.vlanid,
                          llis,
                          ip_mac_pairs)
        self.driver.client.disconnect.assert_called_once()


class TestFOSSWVlanDriverClearVlan(BaseTestFOSSWVlanDriver):
    """Test Fujitsu FOSSW vlan driver for clear vlan"""
    def setUp(self):
        super(TestFOSSWVlanDriverClearVlan, self).setUp()
        self.driver.client = mock.Mock()
        self.driver.client.connect.return_value = None
        self.driver.client.clear_vlan.return_value = None
        self.driver.client.disconnect.return_value = None

    def test_clear_vlan(self):
        ip_mac_pairs = {'00:00:4c:ee:e5:39': '192.168.1.1'}
        self.driver.clear_vlan(LLI['single'], ip_mac_pairs)
        self.driver.client.disconnect.assert_called_once()

    def test_clear_vlan_fail_with_keyerror(self):
        ip_mac_pairs = {'00:00:4c:ee:e5:40': '192.168.1.1'}
        self.assertRaises(
            client.FOSSWClientException,
            self.driver.clear_vlan, LLI['single'], ip_mac_pairs)
        self.driver.client.disconnect.assert_not_called()

    def test_clear_vlan_fail_with_connect_fail(self):
        ip_mac_pairs = {'00:00:4c:ee:e5:39': '192.168.1.1'}
        error = client.FOSSWClientException('clear_vlan')
        self.driver.client.connect.side_effect = error
        self.assertRaises(
            client.FOSSWClientException,
            self.driver.clear_vlan, LLI['single'], ip_mac_pairs)
        self.driver.client.clear_vlan.assert_not_called()
        self.driver.client.disconnect.assert_called_once()


class TestFOSSWVlanDriverClearVlanWithLAG(BaseTestFOSSWVlanDriver):
    """Test Fujitsu FOSSW vlan driver for clear vlan with LAG"""
    def setUp(self):
        super(TestFOSSWVlanDriverClearVlanWithLAG, self).setUp()
        self.driver.client = mock.Mock()
        self.driver.is_valid_mlag = mock.Mock()
        self.driver.clear_vlan = mock.Mock(return_value=None)
        self.ip_macs = {"00:00:4c:ee:e5:39": "192.168.1.1",
                        "00:00:4c:ee:e5:40": "192.168.1.2"}
        self.ports = ','.join(list(sorted(
            [lli['port_id'] for lli in LLI['lag']])))

    def test_lag_normal(self):
        llis = LLI['lag']
        lag_lli = copy.deepcopy(llis[0])
        lag_lli['port_id'] = '3/1'
        self.driver.client.get_lag_port.return_value = "3/1"
        self.assertIsNone(
            self.driver.clear_vlan_with_lag(llis, self.ip_macs))
        self.driver.is_valid_mlag.assert_not_called()
        self.driver.client.connect.assert_called_once_with(
            self.ip_macs[llis[0]['switch_id']])
        self.driver.client.get_lag_port.assert_called_once_with(
            llis[0]['port_id'])
        for lli in llis:
            self.driver.client.leave_from_lag.assert_called_with(
                mock.ANY, '3/1')
        self.driver.clear_vlan.assert_called_once_with([lag_lli], self.ip_macs)
        # Not called because following methods are for MLAG
        self.driver.client.leave_from_vpc.assert_not_called()
        self.driver.client.get_vpcid.assert_not_called()
        self.driver.client.disconnect.assert_called_once()

    def test_lagport_already_cleared(self):
        llis = LLI['lag']
        self.driver.client.get_lag_port.return_value = None
        self.assertIsNone(
            self.driver.clear_vlan_with_lag(llis, self.ip_macs))

        self.driver.client.connect.assert_called_once_with(
            self.ip_macs[llis[0]['switch_id']])
        self.driver.client.get_lag_port.assert_called_once_with(
            llis[0]['port_id'])
        self.driver.client.disconnect.assert_called_once()
        # Not called because following methods are for MLAG
        self.driver.client.leave_from_lag.assert_not_called()
        self.driver.clear_vlan.assert_not_called()
        self.driver.client.leave_from_vpc.assert_not_called()
        self.driver.client.get_vpcid.assert_not_called()

    def test_mlag(self):
        llis = LLI['mlag']
        self.driver.is_valid_mlag.return_value = True
        self.driver.client.get_lag_port.return_value = "3/1"
        self.driver.client.get_vpcid.return_value = "1"
        self.assertIsNone(self.driver.clear_vlan_with_lag(llis, self.ip_macs))
        self.driver.is_valid_mlag.assert_called_once_with(
            list(sorted(list(self.ip_macs.keys()))), self.ip_macs)
        self.driver.client.leave_from_vpc.assert_called_with("3/1", "1")
        self.assertEqual(2, self.driver.client.leave_from_vpc.call_count)
        lag_p = [{
            'switch_id': llis[0]['switch_id'],
            'port_id': '3/1',
            'switch_info': 'ET-7648BRA-FOS'}, {
            'switch_id': llis[1]['switch_id'],
            'port_id': '3/1',
            'switch_info': 'ET-7648BRA-FOS',
        }]
        for idx, arg in enumerate(self.driver.clear_vlan.call_args_list):
            self.assertEqual([lag_p[idx]], arg[0][0])
            self.assertEqual(self.ip_macs, arg[0][1])

    def test_mlag_is_invalid(self):
        llis = LLI['mlag']
        self.driver.is_valid_mlag.return_value = False
        self.assertRaises(client.FOSSWClientException,
                          self.driver.clear_vlan_with_lag, llis, self.ip_macs)

    def test_mlag_vpcid_already_cleared(self):
        llis = LLI['mlag']
        self.driver.is_valid_mlag.return_value = True
        self.driver.client.get_lag_port.return_value = "3/1"
        self.driver.client.get_vpcid.return_value = None
        self.assertIsNone(self.driver.clear_vlan_with_lag(llis, self.ip_macs))
        self.assertEqual(2, self.driver.client.disconnect.call_count)

    def test_lagport_and_vpcid_already_cleared(self):
        llis = LLI['mlag']
        self.driver.is_valid_mlag.return_value = True
        self.driver.client.get_lag_port.return_value = None
        self.driver.client.get_vpcid.return_value = None
        self.driver.clear_vlan_with_lag(llis, self.ip_macs)

        self.assertEqual(2, self.driver.client.connect.call_count)
        self.assertEqual(2, self.driver.client.get_lag_port.call_count)
        self.driver.client.leave_from_lag.assert_not_called()
        self.driver.client.leave_from_vpc.assert_not_called()
        self.driver.client.get_vpcid.assert_not_called()
        self.assertEqual(2, self.driver.client.disconnect.call_count)



# Copyright 2015-2016 FUJITSU LIMITED
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


"""Fujitsu C-Fabric Driver implements CLI over telnet for
Neutron network life-cycle management.
"""

import copy
import re
import select
import time

import eventlet
telnetlib = eventlet.import_patched('telnetlib')

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils

from networking_fujitsu._i18n import _
from networking_fujitsu._i18n import _LE
from networking_fujitsu._i18n import _LW
from networking_fujitsu.ml2.common import utils as fj_util
from neutron.common import utils
from neutron.plugins.ml2.common import exceptions as ml2_exc


LOG = logging.getLogger(__name__)
TELNET_PORT = 23

_EP = 'endpoint'
_PORT_MODE = 'cfab port-mode'
_IFGROUP = 'ifgroup'
_LAG = 'linkaggregation'
_LOCK_NAME = 'fujitsu-cfab'
_TIMEOUT = 30
_TIMEOUT_LOGIN = 5
_WAIT_FOR_BUSY = 3.0
_CRLF_RE = re.compile(r"\r\n", re.MULTILINE)
_PROMPT_LOGIN = "Login: "
_PROMPT_PASS = "Password: "
_PROMPT_ADMIN = "# "
_PROMPT_CONFIG = "(config)# "
_RETRY_PROMPTS_RE = re.compile(
    r"The system is busy|too many sessions")
_PROMPTS_RE = re.compile(
    r"(((\(config(-if)?\))?#)|((?<!<ERROR)>)|((Login|Password):)) $")
_ADMIN_PROMPTS_RE = re.compile(r"((\(config(-if)?\))?#) $")

_MODE_LOGIN = 'login'
_MODE_ADMIN = 'admin'
_MODE_CONFIG = 'config'
_MODE_CONFIG_IF = 'config-if'
_MODE_USER = 'user'

_PAGER_ENABLE_RE = re.compile(r"^pager\s+enable", re.MULTILINE)
_INDEX_RE = re.compile(r"^(\d+)\s+", re.MULTILINE)
_VFAB_PPROFILE_RE = re.compile(
    r"^vfab\s+(default|\d+)\s+pprofile\s+(\d+)\s+"
    r"vsiid\s+(?:mac|uuid)\s+\S+\s+(\S+)", re.MULTILINE)
_IFGROUP_RE = re.compile(r"^ifgroup\s+(\d+)\s+(?:ether|linkaggregation)\s+",
                         re.MULTILINE)
_LAG_RE = re.compile(
    r"^(?:linkaggregation\s+(?:\d+)\s(\d+)\s+)|"
    r"^interface\s+\d/\d/\d/\d$\n(?:(?:\s+.+\n)"
    r"*(?:\s+type\s+linkaggregation\s+(\d+))(?:\s+.+\n)*)\s+exit$",
    re.MULTILINE)
_VFAB_VLAN = r'^vfab\s+{v}\s+vlan\s+{vlan}\s+endpoint\s+{vlan_type}\s+(\d.*)'
_VFAB_VLANS = r'^vfab\s+{v}\s+vlan\s+(\d.*)\s+endpoint\s+{vlan_type}\s+(\d.*)'
_IFGROUP_BOUNDARY = re.compile(r'(\d+)-(\d+)')
_PPROFILE_INDICES = frozenset(range(0, 4096))
_IFGROUP_INDICES = frozenset(range(0, 4095))
_LAG_INDICES = frozenset(range(1, 200))
_INDICES = {_IFGROUP: _IFGROUP_INDICES, _LAG: _LAG_INDICES}
_REG = {_IFGROUP: _IFGROUP_RE, _LAG: _LAG_RE}


class _CFABManager(object):
    """C-Fabric CLI manager.

    Manages C-Fabric using a single telnet connection.
    """

    def __init__(self):
        self._address = None
        self._username = None
        self._password = None
        self._telnet = None
        self._timeout = _TIMEOUT
        self.save_config = True
        self._max_retry = _TIMEOUT / _WAIT_FOR_BUSY
        self._retry_count = 0

    def connect(self, address, username, password):
        """Connect via TELNET and initialize the CLI session."""

        # Use the persisted TELNET connection
        if ((self._telnet and address == self._address and
             username == self._username and password == self._password)):
            return

        # Connect using saved information
        self._address = address
        self._username = username
        self._password = password
        self._reconnect()

    def get_candidate_config(self, prefix=None):
        """Get running-config of the switch."""

        terminal = self._execute("show terminal")
        match = _PAGER_ENABLE_RE.search(terminal)
        if match:
            self._execute("terminal pager disable")
        res = self._get_candidate_config_no_pager_control(prefix)
        if match:
            self._execute("terminal pager enable")
        return res

    def _get_candidate_config_no_pager_control(self, prefix=None):
        """Get running-config of the switch without pager control."""

        current = self._get_mode()
        if current not in (_MODE_ADMIN, _MODE_CONFIG, _MODE_CONFIG_IF):
            self._reconnect()
        cmd = "show candidate-config"
        if prefix:
            cmd = " ".join([cmd, prefix])
        return self._execute(cmd)

    def get_running_config(self, prefix=None):
        """Get running-config of the switch."""

        terminal = self._execute("show terminal")
        match = _PAGER_ENABLE_RE.search(terminal)
        if match:
            self._execute("terminal pager disable")
        res = self._get_running_config_no_pager_control(prefix)
        if match:
            self._execute("terminal pager enable")
        return res

    def _get_running_config_no_pager_control(self, prefix=None):
        """Get running-config of the switch without pager control."""

        current = self._get_mode()
        if current not in (_MODE_ADMIN, _MODE_CONFIG, _MODE_CONFIG_IF):
            self._reconnect()
        cmd = "show running-config"
        if prefix:
            cmd = " ".join([cmd, prefix])
        return self._execute(cmd)

    def configure(self, cmds, commit=True):
        """Configures the switch."""

        self._set_mode_config()
        for cmd in cmds:
            self._execute(cmd)
        if commit:
            self._execute("commit")
            if self.save_config:
                self._execute("save")

    def _set_mode_config(self):
        """Sets the configure mode of the switch."""

        current = self._get_mode()
        if current == _MODE_CONFIG:
            return
        if current == _MODE_ADMIN:
            self._write("configure\n")
            prompt = self._read_until(_PROMPT_CONFIG)
        elif current == _MODE_CONFIG_IF:
            self._write("exit\n")
            prompt = self._read_until(_PROMPT_CONFIG)
        else:
            self._reconnect()
            self._write("configure\n")
            prompt = self._read_until(_PROMPT_CONFIG)
        if prompt.find(_PROMPT_CONFIG) < 0:
            LOG.error(_LE("Failed to set configure mode."))
            raise ml2_exc.MechanismDriverError(method="_set_mode_config")

    def _get_mode(self):
        """Gets the current mode of the switch."""

        self._read_eager()
        self._write("\n")
        idx, match, s = self._expect([_PROMPTS_RE])
        if idx < 0 or match is None:
            LOG.error(_LE("Unexpected response from switch: %s"), s)
            raise ml2_exc.MechanismDriverError(method="_get_mode")
        return _get_mode_from_match(match)

    def _execute(self, cmd):
        """Execute the command on the switch."""

        self._write(cmd + "\n")
        idx, match, s = self._expect([_PROMPTS_RE])
        if idx < 0 or match is None:
            LOG.error(_LE("Unexpected response from switch: %s"), s)
            raise ml2_exc.MechanismDriverError(method="_execute")
        if s.find("<ERROR>") >= 0:
            LOG.error(_LE("Error is returned from switch: %s"), s)
            raise ml2_exc.MechanismDriverError(method="_execute")
        s = _CRLF_RE.sub(r"\n", s)
        # Remove command and prompt
        return s[s.find("\n") + 1:s.rfind("\n") + 1]

    def _reconnect(self):
        """Re-connect and initialize the CLI session."""

        # Close the old connection
        self.close_session()
        # Open new TELNET connection
        try:
            self._telnet = telnetlib.Telnet(
                host=self._address, port=TELNET_PORT, timeout=self._timeout)
        except (OSError, EnvironmentError):
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("Connect failed to switch"))
        try:
            prompt = ""
            prompt = self._telnet.read_until(_PROMPT_LOGIN, _TIMEOUT_LOGIN)
            prompt.index(_PROMPT_LOGIN)
            self._telnet.write(self._username + "\n")
            prompt = self._telnet.read_until(_PROMPT_PASS, _TIMEOUT_LOGIN)
            prompt.index(_PROMPT_PASS)
            self._telnet.write(self._password + "\n")
            prompt = self._telnet.read_until(_PROMPT_ADMIN, _TIMEOUT_LOGIN)
            prompt.index(_PROMPT_ADMIN)
        except (EOFError, OSError, EnvironmentError, ValueError):
            if _RETRY_PROMPTS_RE.search(prompt.strip()):
                # Wait 3 seconds
                if self._retry_count < self._max_retry:
                    LOG.warning(_LW("Wait %(sec)s and retry. cause=%(cause)s"),
                                dict(sec=_WAIT_FOR_BUSY, cause=prompt))
                    self._retry_count += 1
                    time.sleep(_WAIT_FOR_BUSY)
                    self._reconnect()
                    return
                with excutils.save_and_reraise_exception():
                    self.close_session()
                    self._retry_count = 0
                    LOG.exception(_LE("Number of retry times has reached."))
            else:
                with excutils.save_and_reraise_exception():
                    self.close_session()
                    self._retry_count = 0
                    LOG.exception(_LE("Login failed to switch.(%s)"), prompt)

        self._retry_count = 0
        LOG.debug("Connect success to address %(address)s:%(telnet_port)s",
                  dict(address=self._address, telnet_port=TELNET_PORT))

    def close_session(self):
        """Close TELNET session."""

        if self._telnet:
            self._telnet.close()
            self._telnet = None

    def _write(self, buffer):
        """Write a string to the switch."""

        if not self._telnet and self._address:
            self.connect(self._address, self._username, self._password)
        try:
            self._telnet.write(buffer)
        except (OSError, EnvironmentError):
            self.close_session()
            try:
                self._reconnect()
                self._telnet.write(buffer)
            except (OSError, EnvironmentError):
                with excutils.save_and_reraise_exception():
                    LOG.exception(_LE("Write failed to switch"))

    def _read_eager(self):
        """Read readily available data."""
        if not self._telnet and self._address:
            self.connect(self._address, self._username, self._password)
        try:
            return self._telnet.read_eager()
        except (EOFError, OSError, EnvironmentError):
            self.close_session()
            try:
                self._reconnect()
                return self._telnet.read_eager()
            except (EOFError, OSError, EnvironmentError):
                with excutils.save_and_reraise_exception():
                    LOG.exception(_LE("Read failed from switch"))

    def _read_until(self, match):
        """Read until a given string is encountered or until timeout."""

        if not self._telnet and self._address:
            self.connect(self._address, self._username, self._password)
        try:
            return self._telnet.read_until(match, self._timeout)
        except (EOFError, OSError, EnvironmentError):
            self.close_session()
            try:
                self._reconnect()
                return self._telnet.read_until(match, self._timeout)
            except (EOFError, OSError, EnvironmentError):
                with excutils.save_and_reraise_exception():
                    LOG.exception(_LE("Read failed from switch"))

    def _expect(self, res):
        """Read until one from a list of a regular expressions matches."""

        if not self._telnet and self._address:
            self.connect(self._address, self._username, self._password)
        try:
            return self._telnet.expect(res, timeout=self._timeout)
        except (EOFError, OSError, EnvironmentError, select.error):
            self.close_session()
            try:
                self._reconnect()
                return self._telnet.expect(res, timeout=self._timeout)
            except (EOFError, OSError, EnvironmentError, select.error):
                with excutils.save_and_reraise_exception():
                    LOG.exception(_LE("Read failed from switch"))


def _get_mode_from_match(match):
    """Determines the mode from the match object."""

    mode = None
    if match.group(4):
        mode = _MODE_CONFIG_IF
    elif match.group(3):
        mode = _MODE_CONFIG
    elif match.group(2):
        mode = _MODE_ADMIN
    elif match.group(5):
        mode = _MODE_USER
    elif match.group(6):
        mode = _MODE_LOGIN
    return mode

CFAB_MANAGER = _CFABManager()


class CFABdriver(object):
    """C-Fabric CLI interface driver for Neutron network.

    Handles life-cycle management of Neutron network (leverages AMPP on C-Fab)
    """

    def __init__(self, conf=None):
        if conf:
            self._conf = conf
        else:
            self._conf = cfg.CONF
        self._share_pprofile = self._conf.fujitsu_cfab.share_pprofile
        self._pprofile_prefix = self._conf.fujitsu_cfab.pprofile_prefix
        self._validate_pprofile_prefix()
        self._pprofile_name = "{prefix}{pid}"
        self._save_config = self._conf.fujitsu_cfab.save_config
        self.mgr = CFAB_MANAGER
        self.mgr.save_config = self._save_config

    def _validate_pprofile_prefix(self):
        if len(self._pprofile_prefix) > (28 if self._share_pprofile else 15):
            raise ValueError(_("pprofile_prefix is too long."))
        if (self._pprofile_prefix.find('"') >= 0 or
                self._pprofile_prefix.find("|") >= 0 or
                self._pprofile_prefix.find("?") >= 0):
            raise ValueError(_("pprofile_prefix contains illegal character."))

    def _create_ifgroup(self, ifgroup_id, ports, lag_id=None, commit=False):
        """Create ifgroup."""

        common_def = "ifgroup {if_id} {port_type} "
        if lag_id:
            port_type = _LAG
            com = common_def + "{domain} {lag}"
            cmds = [com.format(if_id=ifgroup_id, port_type=port_type,
                               domain=_get_domain_id(ports), lag=lag_id)]
        else:
            port_type = "ether"
            com = common_def + "{ports}"
            cmds = [com.format(if_id=ifgroup_id, port_type=port_type,
                               ports=ports)]
        self.mgr.configure(cmds, commit=commit)

    def _setup_interfaces(self, ports, definitions, commit=False):
        """Add definitions for specified interfaces(ports)."""

        commands = []
        # interface range {ports}
        commands.append("interface range {ports}".format(ports=ports))
        # type endpoint and cfab port-mode external
        for key in sorted(definitions):
            commands.append("{k} {v}".format(k=key, v=definitions[key]))
        commands.append("exit")
        self.mgr.configure(commands, commit=commit)

    def _setup_vfab_vlan(self, vfab_id, vlanid, ifgroup_id, config, ifg=None,
                         port_type=_EP, vlan_type='untag', commit=False):
        """Setup VFAB VLAN configuration."""

        # Judge the result of cleanup process.
        #     ifg is None
        #         'ifgroup' for vfab vlan hasn't changed at all.
        #     ifg is ''(Empty string)
        #         The definition of vfab vlan has been deleted in cleanup.
        #     ifg is String(Not empty)
        #         'ifgroup' for vfab vlan has been changed in cleanup.
        modified = True if ifg is not None else False
        ifgroups = ifg if modified else _get_ifgroups_of_vfab_vlan(
            vfab_id, vlanid, config)
        # Need to re-define for the vfab vlan.
        if modified and ifg is '':
            ifgroups = None
        if _is_ifgroup_included(ifgroup_id, ifgroups):
            LOG.debug(_("ifgroup %(if_id)s has already configured"
                        "in %(ifgs)s"), dict(if_id=ifgroup_id, ifgs=ifgroups))
        elif ifgroups is None:
            ifgroups = str(ifgroup_id)
        else:
            ifgroups += "," + str(ifgroup_id)

        cmds = [
            "vfab {vfab} vlan {vlan} {port_type} {vlan_type} {ifgroup}".format(
                vfab=vfab_id, vlan=vlanid, port_type=port_type,
                vlan_type=vlan_type, ifgroup=ifgroups)]
        self.mgr.configure(cmds, commit=commit)

    def _setup_vlan(self, vfab_id, vlanid, ports, config, ifg=None,
                    commit=False):
        """Setup untagged VLAN configuration."""

        indices = search_ifgroup_indices(ports, config)
        if not indices:
            ifgroup_id = _get_available_index(_IFGROUP, config)
            if ifgroup_id is None:
                raise ml2_exc.MechanismDriverError(method="_setup_vlan")
            self._create_ifgroup(ifgroup_id, ports)
        else:
            ifgroup_id = indices[0]
        self._setup_interfaces(ports, {'type': _EP, _PORT_MODE: 'external'})
        self._setup_vfab_vlan(vfab_id, vlanid, ifgroup_id, config, ifg=ifg,
                              commit=commit)

    def _setup_vlan_with_lag(self, vfab_id, vlanid, ports, config,
                             ifg=None, commit=False):
        """Setup untagged VLAN and LAG configuration."""

        lag_id = self._setup_lag(ports, config)
        indices = search_ifgroup_indices(ports, config, lag_id=lag_id)
        if not indices:
            ifgroup_id = _get_available_index(_IFGROUP, config)
            if ifgroup_id is None:
                raise ml2_exc.MechanismDriverError(
                    method="_setup_vlan_with_lag")
            self._create_ifgroup(ifgroup_id, ports, lag_id=lag_id)
        else:
            ifgroup_id = indices[0]
        defs = {'type': "{p_type} {lag_id}".format(p_type=_LAG, lag_id=lag_id)}
        self._setup_interfaces(ports, defs)
        self._setup_vfab_vlan(vfab_id, vlanid, ifgroup_id, config, ifg=ifg,
                              commit=commit)

    def _setup_lag(self, ports, config):
        """Setup LAG configuration.

        @param self  CFABdriver's instance
        @param ports  a string of the ports which is separated by ','
        @param config  a string of candidate-config for C-Fabric.
        @return lag_id(int)
        """

        lag_id = _get_available_index(_LAG, config)
        if lag_id is None:
            raise ml2_exc.MechanismDriverError(method="_setup_lag")
        # TODO(yushiro) LAG mode 'static', 'active' or 'passive'
        #               Currently, only support 'active'
        mode_opts = {'type': _EP, _PORT_MODE: 'external', 'mode': 'active'}
        self._configure_lag_mode(_get_domain_id(ports), lag_id, mode_opts)
        return lag_id

    def _configure_lag_mode(self, domain_id, lag_id, mode_opts, commit=False):
        """Add LAG definitions.

        @param self  CFABdriver's instance
        @param domain_id  the string of domain_id to which the port belongs
        @param lag_id  the string of LAG definiton ID
        @param mode_opts  the string of LAG mode options
        @param commit  the boolean whether executes commit or not
        @return None
        """

        commands = []
        lag = "linkaggregation {domain_id} {lag_id}".format(
            domain_id=domain_id, lag_id=lag_id)
        for key in sorted(mode_opts):
            commands.append("{lag} {k} {v}".format(lag=lag, k=key,
                                                   v=mode_opts[key]))
        self.mgr.configure(commands, commit=commit)

    def _clear_vlans(self, vfab_id, ports, config, port_type=_EP,
                     vlan_type='untag', lag_id=None, commit=False):
        """Clear all VLAN definitions with specified ports.

        @param self  CFABdriver's instance
        @param vfab_id  the string of VFAB ID
        @param ports  a string of the ports which is separated by ','
        @param config a string of a candidate-config
        @return None or string of the modified ifgroups separated by ','
        """

        # (yushiro): ifgroup won't delete because it can not determine
        #            whether ifgroup is created by plugin or not.
        indices = search_ifgroup_indices(ports, config, lag_id=lag_id)
        vlan_ifgs = _get_all_vfab_vlans_and_ifgroups(vfab_id, config,
                                                     vlan_type='untag')
        updated_vlan_ifgs = copy.deepcopy(vlan_ifgs)
        cmds = []
        for vlanid in sorted(vlan_ifgs.keys()):
            vlan = str(vlanid)
            is_delete = False
            # Target ifgroup_ids exists but vfab vlan definition does not exist
            # or ifgroup_id doesn't exist but vfab vlan definition exists.
            if None in [vlan_ifgs[vlan]] or not indices:
                LOG.debug(_("ifgroup with %(p)s for VLAN(%(v)s)has already"
                            "deleted. Skip clear_vlan."), dict(p=ports,
                                                               v=vlanid))
                continue
            eliminated = fj_util.eliminate_val(vlan_ifgs[vlan], indices)
            updated_vlan_ifgs[vlan] = eliminated
            # VLAN is configured with the only ifgroup_id
            if not eliminated:
                is_delete = True

            common_def = "vfab {vfab} vlan {vlan} {port_type} {vlan_type}"
            # Delete VFAB VLAN definition
            if is_delete:
                command = "no" + " " + common_def
                cmds.append(command.format(vfab=vfab_id, vlan=vlanid,
                                           port_type=port_type,
                                           vlan_type=vlan_type))
            # Reject ifgroup_id from VFAB VLAN definition
            else:
                command = common_def + " " + "{ifgroup}"
                cmds.append(command.format(vfab=vfab_id, vlan=vlanid,
                                           port_type=port_type,
                                           vlan_type=vlan_type,
                                           ifgroup=eliminated))
        self.mgr.configure(cmds, commit=commit)
        return updated_vlan_ifgs

    def _clear_interfaces(self, ports, commit=False):
        """Clear port type definitions with specified interfaces.

        @param self  CFABdriver's instance
        @param ports  a string of the ports which is separated by ','
        @param ether a boolean to judge ether port or linkaggregation
        @param commit a boolean to judge use "commit" or not
        @return None
        """

        cmds = ["interface range {ports}".format(ports=ports)]
        cmds.append("no type")
        cmds.append("no cfab port-mode")
        cmds.append("exit")
        self.mgr.configure(cmds, commit=commit)

    def _clear_lag(self, vfab_id, lag_id, ports, config, commit=False):
        """Clear linkaggregation definition with specified ports.

        @param self  CFABdriver's instance
        @param vfab_id a string of VFAB ID
        @param lag_id a string of LAG ID
        @param ports a string of the ports which is separated by ','
        @param config a string of candidate-config
        @return None
        """

        prefix = 'no linkaggregation {domain_id} {lag_id}'.format(
                 domain_id=_get_domain_id(ports), lag_id=lag_id)
        cmds = []
        cmds.append('{prefix} {p_mode}'.format(prefix=prefix,
                                               p_mode=_PORT_MODE))
        cmds.append('{prefix} mode'.format(prefix=prefix))
        cmds.append('{prefix} type'.format(prefix=prefix))
        self.mgr.configure(cmds, commit=commit)

    @utils.synchronized(_LOCK_NAME, external=True)
    def setup_vlan(self, address, username, password,
                   vfab_id, vlanid, ports, mac):
        """Setup untagged VLAN with specified ports.

        @param self CFABdriver's instance
        @param address the string of C-Fabric IP address
        @param username the string of C-Fabric username
        @param password the string of C-Fabric password
        @param vfab_id the string of VFAB ID
        @param vlanid the string of VLAN ID
        @param ports string of the ports which is separated by ','
        @param mac  string of the MAC address
        @return None
        """
        try:
            self.mgr.connect(address, username, password)
            config = self.mgr.get_candidate_config()
            modified = self._cleanup_definitions(vfab_id, vlanid, ports,
                                                 config, mac)
            self._setup_vlan(vfab_id, vlanid, ports, config, ifg=modified,
                             commit=True)
            self.mgr.close_session()
        except (EOFError, OSError, EnvironmentError, select.error,
                ml2_exc.MechanismDriverError):
            self.mgr.close_session()
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("CLI error"))

    @utils.synchronized(_LOCK_NAME, external=True)
    def setup_vlan_with_lag(self, address, username, password,
                            vfab_id, vlanid, ports, mac):
        """Setup untagged VLAN and linkaggregation.

        @param self CFABdriver's instance
        @param address the string of C-Fabric IP address
        @param username the string of C-Fabric username
        @param password the string of C-Fabric password
        @param vfab_id the string of VFAB ID
        @param vlanid the string of VLAN ID
        @param ports  string of the ports which is separated by ','
        @param mac  string of the MAC address
        @return None
        """
        try:
            self.mgr.connect(address, username, password)
            config = self.mgr.get_candidate_config()
            modified = self._cleanup_definitions(vfab_id, vlanid, ports,
                                                 config, mac)
            self._setup_vlan_with_lag(vfab_id, vlanid, ports, config,
                                      ifg=modified, commit=True)
            self.mgr.close_session()
        except (EOFError, OSError, EnvironmentError, select.error,
                ml2_exc.MechanismDriverError):
            self.mgr.close_session()
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("CLI error"))

    @utils.synchronized(_LOCK_NAME, external=True)
    def clear_vlan(self, address, username, password,
                   vfab_id, vlanid, ports, mac):
        """Clear untagged VLAN.

        @param self  CFABdriver's instance
        @param address  the string of C-Fabric IP address
        @param username  the string of C-Fabric username
        @param password  the string of C-Fabric password
        @param vfab_id  the string of VFAB ID
        @param vlanid  the string of VLAN ID
        @param ports   string of the ports which is separated by ','
        @param mac  string of the MAC address
        @return None
        """
        try:
            self.mgr.connect(address, username, password)
            config = self.mgr.get_candidate_config()
            self._dissociate_mac_from_port_profile(vfab_id, vlanid, mac,
                                                   config=config,
                                                   do_not_commit=True)
            self._clear_vlans(vfab_id, ports, config)
            self._clear_interfaces(ports, commit=True)
            self.mgr.close_session()
        except (EOFError, OSError, EnvironmentError, select.error,
                ml2_exc.MechanismDriverError):
            self.mgr.close_session()
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("CLI error"))

    def _cleanup_definitions(self, vfab_id, vlanid, ports, config, mac,
                             commit=False):
        """Cleanup existing LAG/VLAN/interface definitions.

        @param self  CFABdriver's instance
        @param vfab_id  the string of VFAB ID
        @param vlanid  the string of VLAN ID
        @param ports   string of the ports which is separated by ','
        @param config  string of candidate-config
        @param mac  string of the MAC address
        @param commit  the boolean whether executes commit or not
        @return  A string of modified ifgroups for vfab vlan or None
        """

        self._dissociate_mac_from_port_profile(vfab_id, vlanid, mac,
                                               config=config,
                                               do_not_commit=True)
        lag_id = _get_associated_lag_id(ports, config)
        commit_interface = True if (commit and not lag_id) else False
        self._clear_interfaces(ports, commit=commit_interface)
        if commit_interface:
            LOG.debug(_('Skip clearing VLAN and LAG.'))
            return None
        modified = self._clear_vlans(vfab_id, ports, config, lag_id=lag_id)
        if lag_id:
            LOG.debug(_('Found LAG%s definiton and clear.'), lag_id)
            self._clear_lag(vfab_id, lag_id, ports, config, commit=commit)
        return modified.get(str(vlanid), None)

    @utils.synchronized(_LOCK_NAME, external=True)
    def clear_vlan_with_lag(self, address, username, password,
                            vfab_id, vlanid, ports, mac):
        """Clear untagged VLAN with linkaggregation.

        @param self  CFABdriver's instance
        @param address  the string of C-Fabric IP address
        @param username  the string of C-Fabric username
        @param password  the string of C-Fabric password
        @param vfab_id  the string of VFAB ID
        @param vlanid  the string of VLAN ID
        @param ports   string of the ports which is separated by ','
        @param mac  string of the MAC address
        @return None
        """

        try:
            self.mgr.connect(address, username, password)
            config = self.mgr.get_candidate_config()
            self._cleanup_definitions(vfab_id, vlanid, ports,
                                      config, mac, commit=True)
            self.mgr.close_session()
        except (EOFError, OSError, EnvironmentError, select.error,
                ml2_exc.MechanismDriverError):
            self.mgr.close_session()
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("CLI error"))

    @utils.synchronized(_LOCK_NAME, external=True)
    def associate_mac_to_network(self, address, username, password,
                                 vfab_id, net_id, mac):
        """Associates a MAC address to virtual network.

        @param self  CFABdriver's instance
        @param address  the string of C-Fabric IP address
        @param username  the string of C-Fabric username
        @param password  the string of C-Fabric password
        @param vfab_id  the string of VFAB ID
        @param net_id  the string of VLAN ID(segmentation_id for network)
        @param mac the string of MAC address
        @return None
        """

        try:
            self.mgr.connect(address, username, password)
            self._associate_mac_to_port_profile(vfab_id, net_id, mac)
            self.mgr.close_session()
        except (EOFError, OSError, EnvironmentError, select.error,
                ml2_exc.MechanismDriverError):
            self.mgr.close_session()
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("CLI error"))

    @utils.synchronized(_LOCK_NAME, external=True)
    def dissociate_mac_from_network(self, address, username, password,
                                    vfab_id, net_id, mac):
        """Dissociates a MAC address from virtual network.

        @param self  CFABdriver's instance
        @param address  the string of C-Fabric IP address
        @param username  the string of C-Fabric username
        @param password  the string of C-Fabric password
        @param vfab_id  the string of VFAB ID
        @param net_id  the string of VLAN ID(segmentation_id for network)
        @param mac the string of MAC address
        @return None
        """

        try:
            self.mgr.connect(address, username, password)
            self._dissociate_mac_from_port_profile(vfab_id, net_id, mac)
            self.mgr.close_session()
        except (EOFError, OSError, EnvironmentError, select.error,
                ml2_exc.MechanismDriverError):
            self.mgr.close_session()
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("CLI error"))

    def _create_port_profile(self, vlanid, mac_address, running_config=None,
                             commit=True):
        """Creates a port profile."""

        if running_config is None:
            running_config = self.mgr.get_running_config()
        if self._share_pprofile:
            pprofile = self._get_pprofile(vlanid, mac_address, running_config)
            if pprofile is None:
                pprofile = self._get_new_pprofile(
                    vlanid, mac_address, running_config)
                self._configure_pprofile(pprofile, vlanid, commit)
        else:
            pprofile = self._get_new_pprofile(
                vlanid, mac_address, running_config)
            match = re.search(
                r"^pprofile\s+{pid}\s+vlan\s+tag\s+([0-9,-]+)".format(
                    pid=re.escape(pprofile)), running_config, re.MULTILINE)
            if match:
                if match.group(1) != str(vlanid):
                    LOG.warning(
                        _LW('Override "pprofile %(pid)s vlan tag %(vids)s" '
                            'to "vlan tag %(vlanid)s"'),
                        dict(pid=pprofile, vids=match.group(1),
                             vlanid=vlanid))
                    self._configure_pprofile(pprofile, vlanid, commit)
            else:
                self._configure_pprofile(pprofile, vlanid, commit)
        return pprofile

    def _configure_pprofile(self, profile, vlanid, commit=True):
        """Configures pprofile."""

        self.mgr.configure(
            ["pprofile {pid} vlan tag {vid}".format(
                pid=profile, vid=vlanid)], commit=commit)

    def _get_pprofile(self, vlanid, mac_address, running_config):
        """Gets the name of existing pprofile."""

        if self._share_pprofile:
            pprofile = None
            match = re.search(
                r"^pprofile\s+({prefix}\S+)\s+vlan\s+tag\s+{vid}$".format(
                    prefix=re.escape(self._pprofile_prefix), vid=vlanid),
                running_config, re.MULTILINE)
            if match:
                pprofile = match.group(1)
        else:
            pprofile = self._get_new_pprofile(
                vlanid, mac_address, running_config)
            match = re.search(
                r"^pprofile\s+{pid}\s+vlan\s+tag\s+{vid}$".format(
                    pid=re.escape(pprofile), vid=vlanid),
                running_config, re.MULTILINE)
            if not match:
                pprofile = None
        return pprofile

    def _get_new_pprofile(self, vlanid, mac_address, running_config):
        """Gets a name of new pprofile for the MAC address or the vlan id."""

        if self._share_pprofile:
            used = re.findall(
                r"^pprofile\s+(\S+)\s+", running_config, re.MULTILINE)
            while True:
                pprofile = self._pprofile_name.format(
                    prefix=self._pprofile_prefix, pid=vlanid)
                if pprofile not in used:
                    return pprofile
                vlanid += 1
        else:
            return self._pprofile_name.format(
                prefix=self._pprofile_prefix, pid=mac_address)

    def _associate_mac_to_port_profile(self, vfab_id, vlanid, mac_address):
        """Associates a MAC address to a port profile."""

        running_config = self.mgr.get_running_config()
        pprofile = self._create_port_profile(
            vlanid, mac_address, running_config, commit=False)
        index, profile_name = _search_vfab_pprofile(
            vfab_id, mac_address, running_config)
        if index is None:
            index = _get_available_vfab_pprofile_index(
                vfab_id, running_config)
            if index is None:
                LOG.error(_LE("No unused vfab pprofile index"))
                raise ml2_exc.MechanismDriverError(
                    method="_associate_mac_to_port_profile")
        else:
            if pprofile == profile_name:
                return
            else:
                LOG.warning(
                    _LW('Override "vfab %(vfab_id)s pprofile %(index)d vsiid '
                        'mac %(mac)s %(profile_name)s" to "vsiid mac %(mac)s '
                        '%(pprofile)s"'),
                    dict(vfab_id=vfab_id, index=index, mac=mac_address,
                         profile_name=profile_name, pprofile=pprofile))
        self.mgr.configure(
            ["vfab {vfab_id} pprofile {index} vsiid mac {mac} "
             "{pid}".format(
                 vfab_id=vfab_id, index=index, mac=mac_address, pid=pprofile)])

    def _dissociate_mac_from_port_profile(self, vfab_id, vlanid, mac_address,
                                          config=None, do_not_commit=False):
        """Dissociates a MAC address from a port profile."""

        config = config if config else self.mgr.get_running_config()
        pprofile = self._get_pprofile(vlanid, mac_address, config)
        if pprofile is None:
            return
        index = _get_vfab_pprofile_index(
            vfab_id, pprofile, mac_address, config)
        if index is not None:
            delete_port_profile = True
            for m in _VFAB_PPROFILE_RE.finditer(config):
                if m.group(3) == pprofile and not (
                        m.group(1) == vfab_id and m.group(2) == index):
                    delete_port_profile = False
                    break
            commit = False if do_not_commit else (not delete_port_profile)
            self.mgr.configure(
                ["no vfab {vfab_id} pprofile {index}".format(
                    vfab_id=vfab_id, index=index)], commit=commit)
            if delete_port_profile:
                commit = False if do_not_commit else delete_port_profile
                self.mgr.configure(["no pprofile {pid}".format(pid=pprofile)],
                                   commit=commit)
        else:
            LOG.warning(
                _LW("No corresponding vfab pprofile for %(vid)s, %(mac)s"),
                dict(vid=vlanid, mac=mac_address))


def _search_vfab_pprofile(vfab_id, mac_address, running_config):
    """Search for the vfab pprofile. Returns (index, pprofile) if found."""

    match = re.search(
        r"^vfab\s+{vfab_id}\s+pprofile\s+(\d+)\s+"
        r"vsiid\s+mac\s+{mac}\s+(\S+)$".format(
            vfab_id=vfab_id, mac=mac_address),
        running_config, re.MULTILINE)
    if match:
        return int(match.group(1)), match.group(2)
    else:
        return None, None


def search_ifgroup_indices(ports, candidate_config, lag_id=None):
    """Search ifgroup ids with specified ports."""

    reg = r"^ifgroup\s+(\d+)\s+ether\s+{ports}$".format(ports=ports)
    if lag_id:
        domain_id = _get_domain_id(ports)
        reg = r"^ifgroup\s+(\d+)\s+{if_type}\s+{domain_id}\s+{lag_id}$".format(
            if_type=_LAG, domain_id=domain_id, lag_id=lag_id)
    match = re.findall(reg, candidate_config, re.MULTILINE)
    if match:
        ifgroup_indices = sorted([int(m) for m in match])
        LOG.debug(_('Found ifgroup for port(%(p)s):%(ifg)s'),
                  dict(p=ports, ifg=ifgroup_indices))
        return ifgroup_indices
    return []


def _get_vfab_pprofile_index(vfab_id, pprofile, mac_address, running_config):
    """Gets the index for vfab pprofile."""

    index = None
    match = re.search(
        r"^vfab\s+{vfab_id}\s+pprofile\s+(\d+)\s+"
        r"vsiid\s+mac\s+{mac}\s+{pid}\b".format(
            vfab_id=vfab_id, mac=mac_address, pid=pprofile),
        running_config, re.MULTILINE)
    if match:
        index = match.group(1)
    return index


def _get_ifgroups_of_vfab_vlan(vfab_id, vlanid, config, vlan_type='untag'):
    """Gets ifgroup definitions for specified vfab vlan.

        @param vfab_id  the string of VFAB ID
        @param vlanid  the string of VLAN ID
        @param config the string of candidate-config
        @param vlan_type 'untag(default)' or 'tag'.
        @return the string of ifgroups otherwise None
    """
    match = re.search(
        _VFAB_VLAN.format(v=vfab_id, vlan=vlanid, vlan_type=vlan_type),
        config, re.MULTILINE)
    if match:
        return match.group(1)
    else:
        return None


def _get_all_vfab_vlans_and_ifgroups(vfab_id, config, vlan_type='untag'):
    """Gets all VFAB VLAN definitions and ifgroups.

        @param vfab_id  the string of VFAB ID
        @param config the string of candidate-config
        @param vlan_type 'untag(default)' or 'tag'.
        @return the string of ifgroups otherwise None
    """
    match = re.findall(_VFAB_VLANS.format(v=vfab_id, vlan_type=vlan_type),
                       config, re.MULTILINE)
    result = dict(match)
    if result:
        LOG.debug(_('VFAB(%(vfab)s){VLANID: ifgroups}:%(result)s'),
                  dict(vfab=vfab_id, result=result))
    else:
        LOG.debug(_('VLAN and ifgroups not found.'))
    return result


def _get_available_index(target, config):
    """Gets an available index for specified target resource."""

    indices = _INDICES[target]
    reg = _REG[target]
    available = indices - set([int(''.join(x)) for x in reg.findall(config)])
    if len(available) > 0:
        return sorted(available)[0]
    else:
        LOG.error(_LE("No unused %s index."), target)
        return None


def _get_associated_lag_id(ports, config):
    """Get lag_id which is associated to the ports."""

    ids = []
    interfaces = ports.split(",")
    for port in interfaces:
        match = re.search(
            r'^interface\s+{port}$\n((?:\s+.+\n)'
            r'*(\s+type\s+{port_type}\s+(\d+))(?:\s+.+\n)*)\s+exit$'.format(
                port=port, port_type='linkaggregation'), config, re.MULTILINE)
        if match:
            ids.append(match.group(3))
    lag_ids = list(set(ids))
    if not lag_ids:
        return None
    if len(lag_ids) > 1:
        LOG.warning(
            _LW("Each port%(ports)s has different LAG ids(%(lag_ids)s)"),
            dict(ports=ports, lag_ids=lag_ids))
    LOG.debug(_("Associated LAG%(lag_ids)s with interfaces:%(ports)s"),
              dict(lag_ids=lag_ids, ports=ports))
    return sorted(lag_ids)[0]


def _get_available_vfab_pprofile_index(vfab_id, running_config):
    """Gets an available index for vfab pprofile."""

    available = _PPROFILE_INDICES - set(
        [int(x) for x in
         re.findall(
             r"^vfab\s+{vfab_id}\s+pprofile\s+(\d+)\s+".format(
                 vfab_id=vfab_id), running_config, re.MULTILINE)])
    if len(available) > 0:
        return sorted(available)[0]
    else:
        return None


def _get_domain_id(physical_ports):
    """Get CFAB domain_id from port."""

    # TODO(yushiro) Consider how to get domain_id
    return physical_ports[0]


def _is_ifgroup_included(ifgroup_id, ifgroups):
    """Judge a specified ifgorup is included into ifgroups or not."""

    if ifgroups is None:
        return False
    if str(ifgroup_id) in ifgroups:
        return True
    match = _IFGROUP_BOUNDARY.findall(ifgroups)
    if match:
        for ifg in match:
            if (int(ifg[0]) <= ifgroup_id and ifgroup_id <= int(ifg[1])):
                return True
    return False



# Copyright 2017 FUJITSU LIMITED
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#


"""Implementation of Fujitsu ML2 Mechanism driver for ML2 Plugin."""

from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_utils import importutils

from networking_fujitsu._i18n import _
from networking_fujitsu._i18n import _LE
from networking_fujitsu._i18n import _LI
from networking_fujitsu._i18n import _LW
from networking_fujitsu.ml2.common import utils

from neutron_lib.api.definitions import portbindings
from neutron_lib import constants

from neutron.plugins.ml2.common import exceptions as ml2_exc
from neutron.plugins.ml2 import driver_api


LOG = logging.getLogger(__name__)
DRIVER = 'networking_fujitsu.ml2.'
FOSSW_VLAN_DRIVER = DRIVER + 'fossw.fossw_vlandriver.FOSSWVlanDriver'
FOSSW_VXLAN_DRIVER = DRIVER + 'fossw.fossw_vxlandriver.FOSSWVxlanDriver'
DEFAULT_VLAN = 1

_SUPPORTED_NET_TYPES = ['vlan', 'vxlan']

ML2_FUJITSU_GROUP = "fujitsu_fossw"
ML2_FUJITSU = [
    cfg.ListOpt(
        'fossw_ips', default='[]',
        help=_("List of IP addresses of all FOS switches.")),
    cfg.StrOpt(
        'username', default='',
        help=_("Username of FOS switches.")),
    cfg.StrOpt(
        'password', default='',
        help=_("Password of FOS switches.")),
    cfg.IntOpt(
        'port', default=22,
        help=_("The port number which is used for SSH connection.")),
    cfg.IntOpt(
        'timeout', default=30,
        help=_("The timeout of SSH connection.")),
    cfg.IntOpt(
        'udp_dest_port', default=4789,
        help=_("The port number of VXLAN UDP destionation on the FOS "
               "switch. All VXLANs on the switches use this UDP port as the "
               "UDP destination port in the UDP header when encapsulating.")),
    cfg.IntOpt(
        'ovsdb_vlanid_range_min', default=2,
        help=_("The minimum VLAN ID in the range that is used for binding "
               "VNI and physical port. The range of 78 VLAN IDs (starts from "
               "this value) will be reserved. Default is 2 (VLAN ID from 2 "
               "to 79 will be reserved).")),
    cfg.IntOpt(
        'ovsdb_port', default=6640,
        help=_("The port number which OVSDB server on the FOS switches "
               "lisetn."))]

cfg.CONF.register_opts(ML2_FUJITSU, ML2_FUJITSU_GROUP)


class FOSSWMechanismDriver(driver_api.MechanismDriver):
    """ML2 Mechanism driver for Fujitsu FOS switches.

    This is the upper layer driver class that interfaces to lower layer (CLI)
    below.
    """

    def __init__(self):
        self._vlan_driver = None
        self._vxlan_driver = None
        self.switches_mac_ip_pair = None
        self.ips = None
        self.initialize()

    def initialize(self):
        """Initialize of variables needed by this class."""

        self.ips = cfg.CONF.fujitsu_fossw.fossw_ips
        LOG.debug(_("loaded switch IP is: %s"), self.ips)
        if len(self.ips) == 0:
            raise cfg.RequiredOptError(
                'fossw_ips', cfg.OptGroup(ML2_FUJITSU_GROUP))
        self.username = cfg.CONF.fujitsu_fossw.username
        self.password = cfg.CONF.fujitsu_fossw.password
        self._vlan_driver = importutils.import_object(
            FOSSW_VLAN_DRIVER, cfg.CONF)
        self._vxlan_driver = importutils.import_object(
            FOSSW_VXLAN_DRIVER, cfg.CONF)
        self.switches_mac_ip_pair = self._vlan_driver.get_switch_mac_ip_pair(
            self.ips)
        LOG.debug(_("Registered FOS switch's IP and MAC pairs: %s"),
                  self.switches_mac_ip_pair)

    @log_helpers.log_method_call
    def create_network_postcommit(self, net_context):
        """Calls setup process for FOS switch.

        Case1: Create VLAN network
                   Create VLAN with specified VLAN ID.
        Case2: Create VXLAN network
                   Create DCVPN with specified VNI.
        Case3: Otherwise
                   Do nothing.

        :param mech_context: context of network
        :type mech_context: NetworkContext

        :returns: None
        :rtype: None

        """

        network = net_context.current
        network_id = network['id']
        tenant_id = network['tenant_id']
        network_type = utils.get_network_type(network)
        seg_id = utils.get_segmentation_id(network)

        if network_type == 'vlan' and seg_id:
            self.create_network_postcommit_vlan(seg_id)
        elif network_type == 'vxlan' and seg_id:
            self.create_network_postcommit_vxlan(network_id, seg_id)

        LOG.info(
            _LI("created network (postcommit): network_id=%(network_id)s "
                "tenant_id=%(tenant_id)s"),
            {'network_id': network_id, 'tenant_id': tenant_id})

    @log_helpers.log_method_call
    def create_network_postcommit_vlan(self, vlanid):
        """Create VLAN to FOS switch.

        :param vlanid: the ID of VLAN to be created
        :type vlanid: string

        :returns: None
        :rtype: None

        """
        method = 'create_network_postcommit'
        for ip in self.ips:
            try:
                self._vlan_driver.create_vlan(ip, vlanid)
            except Exception:
                LOG.exception(_LE("Failed to create vlan(%(vid)s) on switch"
                                  "(%(ip)s)."), {'vid': vlanid, 'ip': ip})
                raise ml2_exc.MechanismDriverError(method=method)

    @log_helpers.log_method_call
    def create_network_postcommit_vxlan(self, net_uuid, seg_id):
        method = 'create_network_postcommit'
        try:
            self._vxlan_driver.create_logical_switch(net_uuid, seg_id)
        except Exception:
            LOG.exception(
                _LE("Failed to create vxlan(%(seg_id)s) on switch(%(ip)s)"),
                {'seg_id': seg_id, 'ip': self.ips})
            raise ml2_exc.MechanismDriverError(method=method)

    @log_helpers.log_method_call
    def delete_network_postcommit(self, net_context):
        """Calls clean process for FOS switch.

        Case1: Baremetal deploy with VLAN network
                   Delete VLAN definition.
        Case2: Baremetal deploy with VXLAN network
                   Delete DCVPN definition.
        Case3: Otherwise:
                   Do nothing.

        :param net_context: context of network
        :type net_context: NetworkContext

        :returns: None
        :rtype: None
        """

        network = net_context.current
        net_id = network['id']
        tenant_id = network['tenant_id']
        network_type = utils.get_network_type(network)
        seg_id = utils.get_segmentation_id(network)

        if network_type == 'vlan' and seg_id:
            self.delete_network_postcommit_vlan(seg_id)
        if network_type == 'vxlan' and seg_id:
            self.delete_network_postcommit_vxlan(net_id)
        LOG.info(
            _LI("Deleted network (postcommit): network_id=%(net)s "
                "tenant_id=%(tenant)s"), {'net': net_id, 'tenant': tenant_id})

    @log_helpers.log_method_call
    def delete_network_postcommit_vlan(self, vlanid):
        """Calls clean vlan process for FOS switch.

        :param vlanid: the ID of VLAN to be deleted
        :type vlanid: string

        :returns: None
        :rtypes: None

        """
        method = 'delete_network_postcommit'
        for ip in self.ips:
            try:
                self._vlan_driver.delete_vlan(ip, vlanid)
            except Exception:
                LOG.exception(
                    _LE("Failed to validate on switch(%(ip)s)."), {'ip': ip})
                raise ml2_exc.MechanismDriverError(method=method)

    @log_helpers.log_method_call
    def delete_network_postcommit_vxlan(self, net_uuid):
        method = 'delete_network_postcommit'
        try:
            self._vxlan_driver.delete_logical_switch(net_uuid)
        except Exception:
            LOG.exception(
                _LE("Failed to validate on switch(%(ip)s)."), {'ip': self.ips})
            raise ml2_exc.MechanismDriverError(method=method)

    @log_helpers.log_method_call
    def delete_port_postcommit(self, mech_context):
        """Calls cleanup process for FOS switch.

        Case1: Baremetal deploy with VLAN
                   Clear VLAN/LAG for specified physical port.
        Case2:
                   Clear VXLAN/LAG for specified physical port.
        Case3: Otherwise
                   Do nothing.

        :param mech_context: context of port
        :type mech_context: PortContext

        :returns: None
        :rtype: None
        """

        port = mech_context.current
        network = mech_context.network
        port_id = port['id']
        network_id = port['network_id']
        tenant_id = port['tenant_id']
        network_type = utils.get_network_type(network)
        if network_type == 'vlan' and utils.is_baremetal(port):
            self.clear_vlan(mech_context)
        if network_type == 'vxlan':
            self.clear_vxlan(mech_context)
        LOG.info(
            _LI("Delete port (postcommit): port_id=%(port_id)s "
                "network_id=%(net_id)s tenant_id=%(tenant_id)s"),
            {'port_id': port_id, 'net_id': network_id, 'tenant_id': tenant_id})

    @log_helpers.log_method_call
    def update_port_postcommit(self, context):
        """Update specified physical port on switch."""
        method = 'update_port_postcommit'
        port = context.current
        network = context.network
        vif_type = port['binding:vif_type']
        network_type = utils.get_network_type(network)
        seg_id = utils.get_segmentation_id(network)
        if utils.is_baremetal(port):
            if (vif_type == 'unbound' and
                    context.original['binding:vif_type'] == 'other'):
                try:
                    # Clear vlan or vxlan from physical port for unbound port
                    if network_type == 'vlan':
                        self.clear_vlan(context, use_original=True)
                    if network_type == 'vxlan':
                        self.clear_vxlan(context, use_original=True)
                except Exception:
                    LOG.exception(_LE("Failed to clear %(network_type)s"
                                      "(%(seg_id)s)."),
                                  {'network_type': network_type,
                                   'seg_id': seg_id})
                    raise ml2_exc.MechanismDriverError(method=method)
        else:
            # Setup vxlan from specified physical port on switch.
            if network_type != 'vxlan' or vif_type == 'unbound':
                return
            # currently supports only one segment per network
            # Setup VXLAN for port which bound with DHCP, router, and vms.
            try:
                self._vxlan_driver.update_physical_port(
                    seg_id, [], port, self.switches_mac_ip_pair)
            except Exception:
                LOG.exception(_LE("Failed to setup VNI(%s)."), seg_id)
                raise ml2_exc.MechanismDriverError(method=method)

    @log_helpers.log_method_call
    def setup_vlan(self, params):
        """Provision VLAN with specified port(s).

        This method will select driver's method.
        Case1: param['lag'] is True
            This method calls 'setup_vlan_with_lag' and clears VLAN and LAG.
        Case2: param['lag'] is False
            This method calls 'setup_vlan' and setup only VLAN.

        :param params: a dictionary of the return value for
                        get_physical_net_params
        :type params: dictionary

        :returns: None
        :rtype: None
        """

        target = 'setup_vlan_with_lag' if params['lag'] else 'setup_vlan'
        try:
            setup_method = getattr(self._vlan_driver, target)
            LOG.info(
                _LI("Call %(target)s.  params: %(params)s"),
                {'target': target, 'params': params}
            )
            setup_method(
                params['vlanid'],
                params['local_link_info'],
                self.switches_mac_ip_pair,
            )
        except Exception:
            LOG.exception(_LE("Failed to setup vlan(%s)"), params['vlanid'])
            raise ml2_exc.MechanismDriverError(method=target)

    @log_helpers.log_method_call
    def clear_vlan(self, context, use_original=False):
        """Clear VLAN with specified port(s).

        This method will select driver's method.
        Case1: param['lag'] is True
            This method calls 'clear_vlan_with_lag' and clears VLAN and LAG.
        Case2: param['lag'] is False
            This method calls 'clear_vlan' and clears only VLAN.

        :param params: a dictionary of the return value for
                       get_physical_net_params
        :type params: dictionary
        :returns: None
        :rtype: None
        """
        if not validate_baremetal_deploy(context, use_original):
            return
        params = self.get_physical_net_params(context, use_original)
        target = 'clear_vlan_with_lag'
        try:
            call_target = target if params['lag'] else 'clear_vlan'
            clear_method = getattr(self._vlan_driver, call_target)
            LOG.info(
                _LI("call %(target)s. params: %(params)s"),
                {'target': call_target, 'params': params})
            clear_method(params['local_link_info'], self.switches_mac_ip_pair)
        except Exception:
            LOG.exception(_LE("Failed to clear vlan(%s)"), params['vlanid'])
            raise ml2_exc.MechanismDriverError(method=target)

    @log_helpers.log_method_call
    def clear_vxlan(self, context, use_original=False):
        """Clear VXLAN from specified physical port on switch."""
        method = "clear_vxlan"
        port = context.original if use_original else context.current
        network = context.network
        vif_type = port['binding:vif_type']
        if vif_type == 'unbound' or utils.get_network_type(network) == 'flat':
            return
        lli = utils.get_physical_connectivity(port)
        try:
            if utils.is_lag(lli):
                self._vlan_driver.clear_vlan_with_lag(
                    lli, self.switches_mac_ip_pair)
                self._vxlan_driver.reset_physical_port_with_lag(
                    lli, port, self.switches_mac_ip_pair)
            else:
                self._vxlan_driver.reset_physical_port(
                    lli, port, self.switches_mac_ip_pair)
        except Exception:
            seg_id = utils.get_segmentation_id(network)
            LOG.exception(_LE("Failed to clear VNI(%s)."), seg_id)
            raise ml2_exc.MechanismDriverError(method=method)

    @log_helpers.log_method_call
    def get_physical_net_params(self, context, use_original=False):
        """Validate physical network parameters for baremetal deployment.

        Validates network & port params and returns dictionary.
        'local_link_information' is a dictionary from Ironic-port.  This value
        includes as follows:
            'switch_id': A string of switch's MAC address
                         This value is equal to 'chassis_id' from LLDP TLV.
            'port_id': A string of switch interface name.
                         This value is equal to 'port_id' from LLDP TLV.
            'switch_info': A string of switch name.
                         This value is equal to 'system_name' from LLDP TLV.

        :param port: a dictionary of port
        :type port: dictionary
        :param network: a network object
        :type network: NetworkContext

        :returns: A dictionary parameters for baremetal deploy
        :rtype: dictionary
        """
        port = context.original if use_original else context.current
        network = context.network
        vlanid = utils.get_segmentation_id(network)
        local_link_info = utils.get_physical_connectivity(port)
        return {
            "local_link_info": local_link_info,
            "vlanid": vlanid,
            "mac": port['mac_address'],
            "lag": utils.is_lag(local_link_info)
        }

    @log_helpers.log_method_call
    def bind_port(self, context):

        port = context.current
        network = context.network
        if not is_supported(network):
            return
        vnic_type = port['binding:vnic_type']
        LOG.debug("Attempting to bind port %(port)s with vnic_type "
                  "%(vnic_type)s on network %(network)s",
                  {'port': port['id'], 'vnic_type': vnic_type,
                   'network': network.current['id']})

        network_type = utils.get_network_type(network)
        to_bind = False

        if network_type == 'vxlan':
            # vxlan driver handles not only baremetal port, but vm, dhcp,
            # router ports.
            self.setup_vxlan(context)
            to_bind = True
        if utils.is_baremetal(port) and network_type == 'vlan':
            self.setup_vlan(self.get_physical_net_params(context))
            to_bind = True
        if to_bind:
            context.set_binding(context.segments_to_bind[0][driver_api.ID],
                                portbindings.VIF_TYPE_OTHER, {},
                                status=constants.PORT_STATUS_ACTIVE)

    @log_helpers.log_method_call
    def setup_vxlan(self, context):
        """Update VXLAN from specified physical port on switch."""
        # currently supports only one segment per network
        port = context.current
        seg_id = utils.get_segmentation_id(context.network)
        lli = utils.get_physical_connectivity(port)
        req_id = context.network._plugin_context.request_id
        try:
            if utils.is_lag(lli):
                self._vlan_driver.setup_vlan_with_lag(
                    DEFAULT_VLAN, lli, self.switches_mac_ip_pair)
                self._vxlan_driver.update_physical_port_with_lag(
                    seg_id, lli, port, self.switches_mac_ip_pair, req_id)
            else:
                self._vxlan_driver.update_physical_port(
                    seg_id, lli, port, self.switches_mac_ip_pair, req_id)
        except Exception:
            LOG.exception(_LE("Failed to setup VNI(%s)."), seg_id)
            raise ml2_exc.MechanismDriverError(method='setup_vxlan')


def is_supported(network):
    """Validate network parameter(network_type and segmentation_id).

    :param network: a network object
    :type network: NetworkContext

    :returns: True if network_type is supported and segmentation_id is included
              otherwise False
    :rtype: boolean
    """

    net_type = utils.get_network_type(network)
    if net_type not in _SUPPORTED_NET_TYPES:
        LOG.warning(_LW("Network type(%s) is not supported. Skip it."),
                    net_type)
        return False
    return True if utils.get_segmentation_id(network) else False


def validate_baremetal_deploy(port_context, use_original=False):
    """Validate baremetal deploy.

    :param port_context: a PortContext object
    :type port_context: PortContext

    :returns: True if enable to baremetal deploy otherwise False
    :rtype: boolean
    """

    port = port_context.original if use_original else port_context.current
    network = port_context.network
    return utils.is_baremetal(port) and is_supported(network)



# Copyright 2017 FUJITSU LIMITED
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import ast
import copy
import random
import socket

from oslo_log import log as logging
from oslo_utils import excutils

from networking_fujitsu._i18n import _
from networking_fujitsu._i18n import _LE
from networking_fujitsu._i18n import _LW
from networking_fujitsu.ml2.common.ovsdb import base_connection
from networking_fujitsu.ml2.common.ovsdb import constants as n_const

LOG = logging.getLogger(__name__)


class OVSDBWriter(base_connection.BaseConnection):
    """Performs transactions to OVSDB server tables."""
    def __init__(self, ovsdb_ip, ovsdb_port):
        super(OVSDBWriter, self).__init__(ovsdb_ip, ovsdb_port)

    def _send_and_receive(self, query, operation_id, rcv_required):
        # The self.send(query) never fails.
        # Either it returns true, or there will be a Exception there.
        # However, none of the business here.
        self.send(query)
        if rcv_required:
            self._get_reply(operation_id)

    def _get_reply(self, operation_id):
        count = 0
        while count < n_const.MAX_RETRIES:
            response = self._recv_data()
            LOG.debug(_("Response from OVSDB server = %s"), str(response))
            if response:
                try:
                    self.response = ast.literal_eval(
                        response.replace(':null', ':None'))
                    if (self.response.get('method', None) != "echo" and
                            self._process_response(operation_id)):
                        return True
                except Exception as ex:
                    with excutils.save_and_reraise_exception():
                        LOG.exception(_LE("Exception while receiving the "
                                          "response for the write request: "
                                          "[%s]"), ex)

            count += 1
        with excutils.save_and_reraise_exception():
            LOG.exception(_LE("Could not obtain response from the OVSDB "
                              "server for the request"))

    def _recv_data(self):
        chunks = []
        chunks.append(self.cache[0:])
        lc = rc = 0
        prev_char = None
        while True:
            try:
                response = self.socket.recv(n_const.BUFFER_SIZE)
                if response:
                    for i, c in enumerate(response):
                        if c == '{' and not (prev_char and
                                             prev_char == '\\'):
                            lc += 1
                        elif c == '}' and not (prev_char and
                                               prev_char == '\\'):
                            rc += 1
                        if lc == rc and lc is not 0:
                            chunks.append(response[0:i + 1])
                            self.cache = response[i + 1:]
                            message = "".join(chunks)
                            return message
                        prev_char = c
                    chunks.append(response)
                else:
                    LOG.warning(_LW("Did not receive any reply from the OVSDB "
                                    "server"))
                    return
            except (socket.error, socket.timeout):
                LOG.warning(_LW("Did not receive any reply from the OVSDB "
                                "server"))
                return

    def _process_response(self, op_id):
        result = self._response(op_id)
        if not result:
            raise base_connection.OVSDBError(
                message="No operation_id(%s) matched to response." % op_id)
        error = result.get("error", None)
        if error:
            raise base_connection.OVSDBError(
                message="Error from the OVSDB server: %s" % error)
        # Check errors in responses of all the subqueries
        outcomes = result.get("result", None)
        if outcomes:
            for outcome in outcomes:
                error = outcome.get("error", None)
                if error:
                    raise base_connection.OVSDBError(
                        message="Error from the OVSDB server: %s" % error)
        return result

    def _response(self, op_id):
        if self.response['id'] == op_id:
            return copy.deepcopy(self.response)

    def get_sw_ep_info(self, rcv_required=True):
        """Get switch endpoint information.

        :returns: (endpoint_ip, endpoint_hostname)
                  endpoint_ip: tunnel_ips in Physical_Switch table
                  endpoint_hostname: name in Physical_Switch table
        :rtype: string, string
        """

        op_id = str(random.getrandbits(128))
        query = {'method': 'transact',
                 'params': [n_const.OVSDB_SCHEMA_NAME,
                            {'op': 'select',
                             'table': 'Physical_Switch',
                             'where': [],
                             'columns': ['tunnel_ips', 'name']}],
                 'id': op_id}
        LOG.debug(_("get_sw_ep_info: query: %s"), query)
        self._send_and_receive(query, op_id, rcv_required)
        try:
            result_rows = self.response['result'][0]['rows']
            return_data = result_rows[0] if result_rows else None
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("Exception while receiving the "
                                  "response for the write request: "
                                  "[%s]"), ex)
        endpoint_ip = return_data['tunnel_ips'] if return_data else ""
        endpoint_hostname = return_data['name'] if return_data else ""
        return (endpoint_ip, endpoint_hostname)

    def insert_logical_switch(self, tunnel_key, logical_switch_name,
                              rcv_required=True):
        """Insert an entry to Logical_Switch OVSDB table.

        Note that if logical_switch_name has already been inserted,
        constraint violation error will get returned.

        :param tunnel_key: The segment ID of Neutron network.
        :type tunnel_key: integer
        :param logical_switch_name: Neutron network uuid, but delete all '-'
                                    to make sure the length is within 32.
        :type logical_switch_name: string

        :returns: None
        """

        commit_dict = {'op': 'commit', 'durable': True}
        op_id = str(random.getrandbits(128))
        query = {'method': 'transact',
                 'params': [n_const.OVSDB_SCHEMA_NAME,
                            {'op': 'insert',
                             'table': 'Logical_Switch',
                             'row': {'name': logical_switch_name,
                                     'tunnel_key': tunnel_key}},
                            commit_dict],
                 'id': op_id}
        LOG.debug(_("insert_logical_switch: query: %s"), query)
        self._send_and_receive(query, op_id, rcv_required)

    def get_logical_switch_uuid(self, logical_switch_name, rcv_required=True):
        """Return _uuid of row in Logical_Switch table with the given name.

        :param logical_switch_name: Neutron network uuid, but delete all '-'
                                    to make sure the length is within 32.
        :type logical_switch_name: string

        :returns: UUID of ROW in Logical_Switch table.
        :rtype: uuid
        """

        op_id = str(random.getrandbits(128))
        query = {'method': 'transact',
                 'params': [n_const.OVSDB_SCHEMA_NAME,
                            {'op': 'select',
                             'table': 'Logical_Switch',
                             'where': [['name', '==', logical_switch_name]]}],
                 'id': op_id}
        LOG.debug(_("get_logical_switch_uuid: query: %s"), query)
        self._send_and_receive(query, op_id, rcv_required)
        try:
            result_rows = self.response['result'][0]['rows']
            return_data = result_rows[0] if result_rows else None
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("Exception while receiving the "
                                  "response for the write request: "
                                  "[%s]"), ex)
        return return_data['_uuid'][1] if return_data else ""

    def delete_logical_switch(self, logical_switch_uuid, rcv_required=True):
        """Delete an entry from Logical_Switch OVSDB table.

        Even if the logical_switch_name does not exist, no ERROR will occur.

        :param logical_switch_uuid: UUID of ROW in Logical_Switch table.
        :type logical_switch_uuid: uuid

        :returns: None
        """

        commit_dict = {'op': 'commit', 'durable': True}
        op_id = str(random.getrandbits(128))
        query = {'method': 'transact',
                 'params': [n_const.OVSDB_SCHEMA_NAME,
                            {'op': 'delete',
                             'table': 'Mcast_Macs_Local',
                             'where': [['logical_switch', '==',
                                        ['uuid', logical_switch_uuid]]]},
                            {'op': 'delete',
                             'table': 'Logical_Switch',
                             'where': [['_uuid', '==',
                                        ['uuid', logical_switch_uuid]]]},
                            commit_dict],
                 'id': op_id}
        LOG.debug(_("delete_logical_switch: query: %s"), query)
        self._send_and_receive(query, op_id, rcv_required)

    def get_binding_vid(self, logical_switch_uuid, rcv_required=True):
        """Get mapping vlanid of binding-pair in Physical_Port table.

        If the uuid is not binding to any port, returns 0.

        :param logical_switch_uuid: UUID of logical_switch in the binding-pair
                                    mapping.
        :type logical_switch_uuid: uuid

        :returns: vlanid in the binding-pair mapping.
        :rtype: integer
        """

        binding_vid = 0
        op_id = str(random.getrandbits(128))
        query = {'method': 'transact',
                 'params': [n_const.OVSDB_SCHEMA_NAME,
                            {'op': 'select',
                             'table': 'Physical_Port',
                             'where': [['vlan_bindings', '!=', ['map', []]]],
                             'columns': ['vlan_bindings']}],
                 'id': op_id}
        LOG.debug(_("get_binding_vid: query: %s"), query)
        self._send_and_receive(query, op_id, rcv_required)
        binding_list = []
        try:
            binding_list = self.response["result"][0]["rows"]
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("Exception while receiving the "
                                  "response for the write request: "
                                  "[%s]"), ex)

        if binding_list:
            for b_c in range(len(binding_list)):
                if (binding_list[b_c]["vlan_bindings"][1][0][1][1] ==
                        logical_switch_uuid):
                    binding_vid = binding_list[b_c]["vlan_bindings"][1][0][0]
                    break
        return binding_vid

    def update_physical_port(self, port_name, vlanid, logical_switch_uuid,
                             rcv_required=True):
        """Update a ROW in Physical_Port OVSDB table.

        Updates Physical Port's VNI to VLAN binding.
        Which also means to connect a physical port (port_name) to a
        Neutron network (logical_switch_uuid).

        :param port_name: Symbolic name for the port of FOS switch.
                          format: <slot/port>
                          example: 0/2
        :type port_name: string
        :param vlanid: Identifier of physical port and logical switch binding.
                        Must be in range 0 to 4095.
        :type vlanid: integer
        :param logical_switch_uuid: UUID of logical_switch in the binding-pair
                                    mapping.
        :type logical_switch_uuid: uuid

        :returns: None
        """

        commit_dict = {'op': 'commit', 'durable': True}
        op_id = str(random.getrandbits(128))
        query = {'method': 'transact',
                 'params': [n_const.OVSDB_SCHEMA_NAME,
                            {'op': 'update',
                             'table': 'Physical_Port',
                             'where': [['name', '==', port_name]],
                             'row': {
                                 'vlan_bindings': [
                                     'map',
                                     [[vlanid, ['uuid',
                                                logical_switch_uuid]]]]}},
                            commit_dict],
                 'id': op_id}
        LOG.debug(_("update_physical_port: query: %s"), query)
        self._send_and_receive(query, op_id, rcv_required)

    def get_ucast_macs_local(self, port_mac, rcv_required=True):
        """Get ROWs of port_mac in Ucast_Macs_Local table.

        :param port_mac: MAC address of physical port.
        :type port_mac: string

        :returns: list of ROWs information of the given port_mac in
                  Ucast_Macs_Local table.
        :rtype: list
        """

        op_id = str(random.getrandbits(128))
        query = {'method': 'transact',
                 'params': [n_const.OVSDB_SCHEMA_NAME,
                            {'op': 'select',
                             'table': 'Ucast_Macs_Local',
                             'where': [['MAC', '==', port_mac]]}],
                 'id': op_id}
        LOG.debug(_("get_ucast_macs_local: query: %s"), query)
        self._send_and_receive(query, op_id, rcv_required)
        try:
            return_list = self.response['result'][0]['rows']
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("Exception while receiving the "
                                  "response for the write request: "
                                  "[%s]"), ex)
        return return_list if return_list else []

    def delete_ucast_macs_local(self, port_mac, rcv_required=True):
        """Delete ROW(s) from Ucast_Macs_Local table.

        Even if the port_mac does not exist, no ERROR will occur.

        :param port_mac: MAC address of physical port.
        :type port_mac: string

        :returns: None
        """

        commit_dict = {'op': 'commit', 'durable': True}
        op_id = str(random.getrandbits(128))
        query = {'method': 'transact',
                 'params': [n_const.OVSDB_SCHEMA_NAME,
                            {'op': 'delete',
                             'table': 'Ucast_Macs_Local',
                             'where': [['MAC', '==', port_mac]]},
                            commit_dict],
                 'id': op_id}
        LOG.debug(_("delete_ucast_macs_local: query: %s"), query)
        self._send_and_receive(query, op_id, rcv_required)

    def get_physical_locator_uuid(self, dst_ip, rcv_required=True):
        """Get _uuid of ROW in Physical_Locator table.

        :param dst_ip: IP address of physical locator.
        :type dst_ip: string

        :returns: uuid of ROW of the given dst_ip in Physical_Locator table.
        :rtype: uuid
        """

        op_id = str(random.getrandbits(128))
        query = {'method': 'transact',
                 'params': [n_const.OVSDB_SCHEMA_NAME,
                            {'op': 'select',
                             'table': 'Physical_Locator',
                             'where': [['dst_ip', '==', dst_ip]]}],
                 'id': op_id}
        LOG.debug(_("get_physical_locator_uuid: query: %s"), query)
        self._send_and_receive(query, op_id, rcv_required)
        try:
            result_rows = self.response['result'][0]['rows']
            return_data = result_rows[0] if result_rows else None
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("Exception while receiving the "
                                  "response for the write request: "
                                  "[%s]"), ex)
        return return_data['_uuid'][1] if return_data else ""

    def insert_ucast_macs_local(self, logical_switch_uuid, locator_uuid,
                                MAC_value, rcv_required=True):
        """Insert ROW to Ucast_Macs_Local table.

        :param logical_switch_uuid: uuid of logical switch.
        :type dst_ip: uuid
        :param locator_uuid: uuid of physical locator which its dst_ip is the
                             tunnel_ips of physical switch.
        :type locator_uuid: uuid
        :param MAC_value: MAC address of NIC of baremetal server.
        :type MAC_value: string

        :returns: None
        """

        commit_dict = {'op': 'commit', 'durable': True}
        op_id = str(random.getrandbits(128))
        query = {'method': 'transact',
                 'params': [n_const.OVSDB_SCHEMA_NAME,
                            {'op': 'insert',
                             'table': 'Ucast_Macs_Local',
                             'row': {'MAC': MAC_value,
                                     'logical_switch': ['uuid',
                                                        logical_switch_uuid],
                                     'locator': ['uuid', locator_uuid]}},
                            commit_dict],
                 'id': op_id}
        LOG.debug(_("insert_ucast_macs_local: query: %s"), query)
        self._send_and_receive(query, op_id, rcv_required)

    def insert_ucast_macs_local_and_locator(self, logical_switch_uuid,
                                            locator_ip, MAC_value,
                                            rcv_required=True):
        """Insert ROW to Ucast_Macs_Local table.

        :param logical_switch_uuid: uuid of logical switch.
        :type dst_ip: uuid
        :param locator_ip: IP address of physical locator which also is the
                           tunnel_ips of physical switch.
        :type locator_ip: string
        :param MAC_value: MAC address of NIC of baremetal server.
        :type MAC_value: string

        :returns: None
        """

        commit_dict = {'op': 'commit', 'durable': True}
        op_id = str(random.getrandbits(128))
        query = {'method': 'transact',
                 'params': [n_const.OVSDB_SCHEMA_NAME,
                            {'op': 'insert',
                             'table': 'Physical_Locator',
                             'row': {'dst_ip': locator_ip,
                                     'encapsulation_type': 'vxlan_over_ipv4'
                                     }, 'uuid-name': 'RVTEP'},
                            {'op': 'insert',
                             'table': 'Ucast_Macs_Local',
                             'row': {'MAC': MAC_value,
                                     'logical_switch': ['uuid',
                                                        logical_switch_uuid],
                                     'locator': ['named-uuid', 'RVTEP']}},
                            commit_dict],
                 'id': op_id}
        LOG.debug(_("insert_ucast_macs_local: query: %s"), query)
        self._send_and_receive(query, op_id, rcv_required)

    def get_ucast_macs_remote(self, port_mac, rcv_required=True):
        """Get ROWs of port_mac in Ucast_Macs_Remote table.

        :param port_mac: MAC address of VM port.
        :type port_mac: string

        :returns: list of ROWs information of the given port_mac in
                  Ucast_Macs_Remote table.
        :rtype: list
        """

        op_id = str(random.getrandbits(128))
        query = {'method': 'transact',
                 'params': [n_const.OVSDB_SCHEMA_NAME,
                            {'op': 'select',
                             'table': 'Ucast_Macs_Remote',
                             'where': [['MAC', '==', port_mac]]}],
                 'id': op_id}
        LOG.debug(_("get_ucast_macs_remote: query: %s"), query)
        self._send_and_receive(query, op_id, rcv_required)
        try:
            return_list = self.response['result'][0]['rows']
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("Exception while receiving the "
                                  "response for the write request: "
                                  "[%s]"), ex)
        return return_list if return_list else []

    def delete_ucast_macs_remote(self, port_mac, rcv_required=True):
        """Delete ROW(s) from Ucast_Macs_Remote table.

        Even if the port_mac does not exist, no ERROR will occur.

        :param port_mac: MAC address of VM port.
        :type port_mac: string

        :returns: None
        """

        commit_dict = {'op': 'commit', 'durable': True}
        op_id = str(random.getrandbits(128))
        query = {'method': 'transact',
                 'params': [n_const.OVSDB_SCHEMA_NAME,
                            {'op': 'delete',
                             'table': 'Ucast_Macs_Remote',
                             'where': [['MAC', '==', port_mac]]},
                            commit_dict],
                 'id': op_id}
        LOG.debug(_("delete_ucast_macs_remote: query: %s"), query)
        self._send_and_receive(query, op_id, rcv_required)

    def insert_ucast_macs_remote(self, logical_switch_uuid, MAC_value,
                                 ipaddrs, locator_uuid, rcv_required=True):
        """Insert ROW(s) to Ucast_Macs_Remote table.

        There are two cases:
        Case 1:
            A VM port has been created or updated in a Compute node somewhere.
        Case 2:
            A BM port has been created or updated in another FOS switch.

        :param logical_switch_uuid: uuid of logical switch.
        :type logical_switch_uuid: uuid
        :param MAC_value: MAC address of port.
        :type MAC_value: string
        :param ipaddrs: list of fixed ips pairs to MAC_value.
        :type ipaddrs: list
        :param locator_uuid: uuid of physical locator (remote VTEP) where
                             the port is binding to.
        :type locator_uuid: uuid

        :returns: None
        """

        commit_dict = {"op": "commit", "durable": True}
        op_id = str(random.getrandbits(128))
        params = [n_const.OVSDB_SCHEMA_NAME]
        for ipaddr in ipaddrs:
            params.append({"op": "insert",
                           "table": "Ucast_Macs_Remote",
                           "row": {"MAC": MAC_value,
                                   "logical_switch": ["uuid",
                                                      logical_switch_uuid],
                                   "locator": ["uuid", locator_uuid],
                                   "ipaddr": ipaddr}})
        params.append(commit_dict)
        query = {"method": "transact",
                 "params": params,
                 "id": op_id}
        LOG.debug(_("insert_ucast_macs_remote: query: %s"), query)
        self._send_and_receive(query, op_id, rcv_required)

    def insert_ucast_macs_remote_and_locator(self, logical_switch_uuid,
                                             MAC_value, ipaddrs, locator_ip,
                                             rcv_required=True):
        """Insert ROW(s) to Ucast_Macs_Remote table.

        Also insert related locator ROW to Pysical_Locator.
        There are two cases:
        Case 1:
        A VM port has been created or updated in a Compute node somewhere.
        Case 2:
        A BM port has been created or updated in another FOS switch.

        :param logical_switch_uuid: uuid of logical switch.
        :type logical_switch_uuid: uuid
        :param MAC_value: MAC address of port.
        :type MAC_value: string
        :param ipaddrs: list of fixed ips pairs to MAC_value.
        :type ipaddrs: list
        :param locator_IP: IP address of physical locator (remote VTEP) where
                           the port is binding to.
        :type locator_uuid: string

        :returns: None
        """

        commit_dict = {"op": "commit", "durable": True}
        op_id = str(random.getrandbits(128))
        params = [n_const.OVSDB_SCHEMA_NAME,
                  {'op': 'insert',
                   'table': 'Physical_Locator',
                   'row': {'dst_ip': locator_ip,
                           'encapsulation_type': 'vxlan_over_ipv4'
                           }, 'uuid-name': 'RVTEP'}]
        for ipaddr in ipaddrs:
            params.append({'op': 'insert',
                           'table': 'Ucast_Macs_Remote',
                           'row': {'MAC': MAC_value,
                                   'logical_switch': ['uuid',
                                                      logical_switch_uuid],
                                   'locator': ['named-uuid', 'RVTEP'],
                                   'ipaddr': ipaddr}})
        params.append(commit_dict)
        query = {"method": "transact",
                 "params": params,
                 "id": op_id}
        LOG.debug(_("insert_ucast_macs_remote: query: %s"), query)
        self._send_and_receive(query, op_id, rcv_required)

    def reset_physical_port(self, port_name, rcv_required=True):
        """Reset a ROW in Physical_Port table.

        :param port_name: Symbolic name for the port of FOS switch.
                          format: <slot/port>
                          example: 0/2
        :type port_name: string

        :returns: None
        """

        commit_dict = {"op": "commit", "durable": True}
        op_id = str(random.getrandbits(128))
        query = {"method": "transact",
                 "params": [n_const.OVSDB_SCHEMA_NAME,
                            {"op": "update",
                             "table": "Physical_Port",
                             "where": [["name", "==", port_name]],
                             "row": {"vlan_bindings": ["map", []]}},
                            commit_dict],
                 "id": op_id}
        LOG.debug(_("reset_physical_port: query: %s"), query)
        self._send_and_receive(query, op_id, rcv_required)



# Copyright 2017 FUJITSU LIMITED
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from neutron.common import topics
from neutron.plugins.ml2 import managers
from neutron.plugins.ml2 import rpc


class TunnelCaller(object):
    """Common tunnel calls for FOSSW agent."""
    def __init__(self):
        self._construct_rpc_stuff()

    def _construct_rpc_stuff(self):
        self.notifier = rpc.AgentNotifierApi(topics.AGENT)
        self.type_manager = managers.TypeManager()
        self.tunnel_rpc = rpc.RpcCallbacks(self.notifier, self.type_manager)

    def trigger_tunnel_sync(self, context, tunnel_ip):
        """Sends tunnel sync RPC message to the neutron L2 agent.

        :param context: A context of request.
        :type context: A Neutron context object.
        :param tunnel_ip: An endpoint IP address of vxlan.
        :type tunnel_ip: string

        :returns: None
        """

        tunnel_dict = {'tunnel_ip': tunnel_ip, 'tunnel_type': 'vxlan'}
        self.tunnel_rpc.tunnel_sync(context, **tunnel_dict)



# Copyright 2017 FUJITSU LIMITED
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import socket
import time

from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import excutils

from networking_fujitsu._i18n import _
from networking_fujitsu._i18n import _LE
from networking_fujitsu._i18n import _LW
from networking_fujitsu.ml2.common.ovsdb import constants as n_const

from neutron_lib import exceptions

LOG = logging.getLogger(__name__)


class OVSDBError(exceptions.NeutronException):
    message = _("%(message)s")


class BaseConnection(object):
    """Connects to OVSDB server.

       Connects to an ovsdb server on a given host and TCP port.
    """
    def __init__(self, ovsdb_ip, ovsdb_port):
        self.response = {}
        self.cache = ""
        self.ovsdb_ip = ovsdb_ip
        self.ovsdb_port = ovsdb_port
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        retryCount = 0
        while True:
            try:
                self.socket.connect((str(self.ovsdb_ip), int(self.ovsdb_port)))
                break
            except (socket.error, socket.timeout):
                LOG.warning(_LW('Unable to reach OVSDB server %s'),
                            self.ovsdb_ip)
                if retryCount == n_const.MAX_CONNECTION_RETRIES:
                    # Retried for max_connection_retries times.
                    # Give up and return so that it can be tried in
                    # the next periodic interval.
                    with excutils.save_and_reraise_exception(reraise=True):
                        LOG.exception(_LE("Socket error in connecting to "
                                          "the OVSDB server"))
                else:
                    time.sleep(1)
                    retryCount += 1

        # Successfully connected to the socket
        LOG.debug(_("Connected to OVSDB server %(ip)s, with tcp port %(port)s"
                    ), {'ip': self.ovsdb_ip, 'port': self.ovsdb_port})

    def disconnect(self):
        """disconnects the connection from the OVSDB server."""
        self.socket.close()

    def send(self, message):
        """Sends a message to the OVSDB server."""
        retry_count = 0
        while retry_count < n_const.MAX_RETRIES:
            try:
                # The self.socket.send always sends more than 1 byte.
                # Because, at least, query id will be add to the query.
                self.socket.send(jsonutils.dumps(message))
                return True
            except Exception as ex:
                LOG.exception(_LE("Exception [%s] occurred while sending "
                                  "message to the OVSDB server"), ex)
            retry_count += 1

        LOG.exception(_LE("Could not send message to the OVSDB server."))
        self.disconnect()
        return False
